#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <paths.h>				/* For _PATH_DEVNULL */
#include <glib.h>
#include "if_config.h"
#include "config.h"


static const gchar *if_config_get_str_hex_to_uint8(
	const gchar *s,
	guint8 *v_rtn
);
static const gchar *if_config_get_str_hex_gsbf_to_address(
	const gchar *s,
	INAddress *v_rtn
);

static const gchar *if_get_config_parse_line_iterate(
	const gchar *line,
	const gchar *if_name,
	IFConfig *config_buf
);
static void if_get_config_parse_line(
	const gchar *line,
	const gchar *if_name,
	IFConfig *config_buf
);
static gint if_get_config_stream(
	FILE *fp,
	const gchar *if_name,
	IFConfig *config_buf
);

gboolean if_get_state(const gchar *if_name);

gint if_get_config(
	const gchar *if_name,
	IFConfig *config_buf
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))

#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)
#define STRPFX(s,p)     ((((s) != NULL) && ((p) != NULL)) ? \
			(!strncmp((s), (p), strlen(p))) : FALSE)

#define ISBLANK(c)	(((c) == ' ') || ((c) == '\t'))


#ifndef _PATH_DEVNULL
# define _PATH_DEVNULL			"/dev/null"
#endif

#ifndef PROG_IFCONFIG
# warning "PROG_IFCONFIG not defined, unable to determine the interface configuration program's path."
#endif


/*
 *	Gets the 1 byte hex value and seeks s past the value.
 */
static const gchar *if_config_get_str_hex_to_uint8(
	const gchar *s,
	guint8 *v_rtn
)
{
	guint8 v[2];

	/* Seek past any "0x" prefix */
	if(STRPFX(s, "0x"))
		s += strlen("0x");

	/* Check for end of string, end of value, or end of line */
	if((*s == '\0') || ISBLANK(*s) || (*s == '\n'))
	{
		*v_rtn = 0x00;
		return(s);
	}

	/* Get the first character value */
	v[0] = (isdigit((int)*s)) ?
		(guint8)(*s - '0') : (guint8)(toupper(*s) - 'A') + 10;

	/* Seek to the next character */
	s++;

	/* Check for end of string, end of value, or end of line */
	if((*s == '\0') || ISBLANK(*s) || (*s == '\n'))
	{
		*v_rtn = 0x00;
		return(s);
	}

	/* Get the second character value */
	v[1] = (isdigit((int)*s)) ?
		(guint8)(*s - '0') : (guint8)(toupper(*s) - 'A') + 10;

	/* Set the return value */
	*v_rtn = (guint8)((v[0] * 16) + v[1]);

	/* Seek past the value and return */
	return(s + 1);
}

/*
 *	Gets the address value with greatest significant byte first and
 *	seeks s past the value.
 */
static const gchar *if_config_get_str_hex_gsbf_to_address(
	const gchar *s,
	INAddress *v_rtn
)
{
	gint i, n;
	const gchar *sp;

	/* Seek past any "0x" prefix */
	if(STRPFX(s, "0x"))
		s += strlen("0x");

	/* Count the number of characters in the value and divide by
	 * 2 to get the number of bytes specified by the string and
	 * not exceeding the end of the string
	 */
	for(sp = s, n = 0;
	    (*sp != '\0') && !ISBLANK(*sp) && (*sp != '\n');
	    sp++, n++
	);
	n = MIN((n / 2), 4);

	/* Get the address value as the greatest significant byte first
	 * and seek s past the value
	 */
	for(i = n - 1; i >= 0; i--)
		s = if_config_get_str_hex_to_uint8(
			s,
			&v_rtn->u8[i]
		);

	return(s);
}


/*
 *	Gets the next value from the line.
 *
 *	Returns the pointer to the next value or end of line.
 */
static const gchar *if_get_config_parse_line_iterate(
	const gchar *line,
	const gchar *if_name,
	IFConfig *config_buf
)
{
	const gchar *s = line;

#define SEEK_PAST_SPACES(_s_)	{			\
 while(ISBLANK(*(_s_)))					\
  (_s_)++;						\
}
#define SEEK_PAST_WORD(_s_)	{			\
 while((*s != '\0') && !ISBLANK(*(_s_)))		\
  (_s_)++;						\
}
#define GET_ADDRESS_4DOTS(_s_,_address_)	{	\
 gint i;						\
 INAddress *addr = (_address_);				\
 for(i = 0; i < 4; i++) {				\
  addr->u8[i] = (guint8)atoi((const char *)(_s_));	\
  while((*(_s_) != '\0') && !ISBLANK(*(_s_)) &&		\
   (*(_s_) != '\n')					\
  ) {							\
   if(*(_s_) == '.') {					\
    (_s_)++;						\
    break;						\
   }							\
   (_s_)++;						\
  }							\
 }							\
}

	/* Get the value */
	if(STRPFX(s, "Link encap:"))
	{
		gint len;
		const gchar *s_end;

		s += strlen("Link encap:");
		SEEK_PAST_SPACES(s);

		s_end = (const gchar *)strstr((const char *)s, "  ");
		if(s_end == NULL)
			s_end = (const gchar *)strchr((const char *)s, '\n');
		len = (s_end != NULL) ? s_end - s : strlen(s);
		if(len > 0)
		{
			const gint buf_len = sizeof(config_buf->link_encap);
			gchar *buf = config_buf->link_encap;
			(void)strncpy(
				buf,
				s,
				(size_t)buf_len
			);
			if(len < buf_len)
				buf[len] = '\0';
			else
				buf[buf_len - 1] = '\0';

			s += len;
		}
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "HWaddr"))
	{
		gint i;
		IFHardwareAddress *addr = &config_buf->hw_addr;

		s += strlen("HWaddr");
		SEEK_PAST_SPACES(s);

		/* Get the 6 byte hardware address value as the
		 * least significent byte first
		 */
		for(i = 0; i < 6; i++)
		{
			s = if_config_get_str_hex_to_uint8(
				s,
				&addr->u8[i]
			);
			if(*s == ':')
				s++;
		}
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "inet addr:"))
	{
		s += strlen("inet addr:");
		SEEK_PAST_SPACES(s);
		GET_ADDRESS_4DOTS(s, &config_buf->inet_addr);
	}
	else if(STRPFX(s, "Bcast:"))
	{
		s += strlen("Bcast:");
		SEEK_PAST_SPACES(s);
		GET_ADDRESS_4DOTS(s, &config_buf->bcast);
	}
	else if(STRPFX(s, "P-t-P:"))
	{
		s += strlen("P-t-P:");
		SEEK_PAST_SPACES(s);
		GET_ADDRESS_4DOTS(s, &config_buf->ptp);
	}
	else if(STRPFX(s, "Mask:"))
	{
		s += strlen("Mask:");
		SEEK_PAST_SPACES(s);
		GET_ADDRESS_4DOTS(s, &config_buf->mask);
	}
	else if(STRPFX(s, "MTU:"))
	{
		s += strlen("MTU:");
		SEEK_PAST_SPACES(s);
		config_buf->mtu = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "Metric:"))
	{
		s += strlen("Metric:");
		SEEK_PAST_SPACES(s);
		config_buf->metric = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "RX packets:"))
	{
		s += strlen("RX packets:");
		SEEK_PAST_SPACES(s);
		config_buf->rx_packets = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		/* Need to get subsequent values since they are only
		 * on this line and have no other identifyable means
		 * without the prefix
		 */
		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->rx_errors = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->rx_dropped = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->rx_overruns = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->rx_frame = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "TX packets:"))
	{
		s += strlen("TX packets:");
		SEEK_PAST_SPACES(s);
		config_buf->tx_packets = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		/* Need to get subsequent values since they are only
		 * on this line and have no other identifyable means
		 * without the prefix
		 */
		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->tx_errors = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->tx_dropped = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->tx_overruns = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		while((*s != '\0') && !isdigit((int)*s))
			s++;
		config_buf->tx_carrier = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "collisions:"))
	{
		s += strlen("collisions:");
		SEEK_PAST_SPACES(s);
		config_buf->collisions = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "txqueuelen:"))
	{
		s += strlen("txqueuelen:");
		SEEK_PAST_SPACES(s);
		config_buf->tx_queue_len = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "RX bytes:"))
	{
		s += strlen("RX bytes:");
		SEEK_PAST_SPACES(s);
		config_buf->rx_bytes = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "TX bytes:"))
	{
		s += strlen("TX bytes:");
		SEEK_PAST_SPACES(s);
		config_buf->tx_bytes = (gulong)atol((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "Interrupt:"))
	{
		s += strlen("Interrupt:");
		SEEK_PAST_SPACES(s);
		config_buf->interrupt = (gint)atoi((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}
	else if(STRPFX(s, "Base address:"))
	{
		s += strlen("Base address:");
		SEEK_PAST_SPACES(s);
		if_config_get_str_hex_gsbf_to_address(
			s,
			&config_buf->base_address
		);
		SEEK_PAST_SPACES(s);
	}
	/* Flags - these need to be parsed last since they have the
	 * potential to be confused with partial parameter names
	 */
	else if(STRPFX(s, "UP"))
	{
		s += strlen("UP");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_UP;
	}
	else if(STRPFX(s, "LOOPBACK"))
	{
		s += strlen("LOOPBACK");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_LOOPBACK;
	}
	else if(STRPFX(s, "BROADCAST"))
	{
		s += strlen("BROADCAST");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_BROADCAST;
	}
	else if(STRPFX(s, "POINTTOPOINT"))
	{
		s += strlen("POINTTOPOINT");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_POINTTOPOINT;
	}
	/* For ifconfig typo */
	else if(STRPFX(s, "POINTOPOINT"))
	{
		s += strlen("POINTOPOINT");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_POINTTOPOINT;
	}
	else if(STRPFX(s, "NOTRAILERS"))
	{
		s += strlen("NOTRAILERS");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_NOTRAILERS;
	}
	else if(STRPFX(s, "RUNNING"))
	{
		s += strlen("RUNNING");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_RUNNING;
	}
	else if(STRPFX(s, "NOARP"))
	{
		s += strlen("NOARP");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_NOARP;
	}
	else if(STRPFX(s, "MULTICAST"))
	{
		s += strlen("MULTICAST");
		SEEK_PAST_SPACES(s);
		config_buf->flags |= IF_CONFIG_FLAG_MULTICAST;
	}
	else
	{
		/* Unsupported word */
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);
	}

#undef GET_ADDRESS_4DOTS
#undef SEEK_PAST_WORD
#undef SEEK_PAST_SPACES

	return(s);
}

/*
 *	Gets all the values from the line.
 */
static void if_get_config_parse_line(
	const gchar *line,
	const gchar *if_name,
	IFConfig *config_buf
)
{
	const gchar *s = line;
	while((*s != '\0') && (*s != '\n'))
	{
		s = if_get_config_parse_line_iterate(
			s,
			if_name,
			config_buf
		);
	}
}

/*
 *	Gets all of the interface's configuration values from the
 *	stream.
 */
static gint if_get_config_stream(
	FILE *fp,
	const gchar *if_name,
	IFConfig *config_buf
)
{
	gboolean	got_interface = FALSE,
			in_context = FALSE;
	gchar		*s,
			buf[1024];

#if 1
	/* Read each line from the stream and get the interface's
	 * configuration values from it
	 */
	while(fgets(buf, sizeof(buf), fp) != NULL)
	{
		s = buf;

		/* Did we receive a line with the interface's name? */
		if(in_context)
		{
			/* If the line starts with a non blank
			 * character then it implies that we have
			 * reached the end of the values for the
			 * specified interface
			 */
			if(!ISBLANK(*s))
			{
/*				in_context = FALSE; */
				break;
			}

			/* Seek the spaces */
			while(ISBLANK(*s))
				s++;

			/* Get the values on this line */
			if_get_config_parse_line(
				s,
				if_name,
				config_buf
			);
		}
		else
		{
			/* Not in context
			 *
			 * If this line does not start with a blank
			 * character then it implies it is the start
			 * of the values for an interface
			 */
			if(!ISBLANK(*s))
			{
				/* Look for the specified interface's
				 * name on this line */
				if(STRPFX(s, if_name))
				{
					in_context = TRUE;
					got_interface = TRUE;

					/* Seek past the interface's name */
					s += strlen((const char *)if_name);

					/* Seek past any spaces */
					while(ISBLANK(*s))
						s++;

					/* Get the subsequent values
					 * on this line
					 */
					if_get_config_parse_line(
						s,
						if_name,
						config_buf
					);
				}
			}
		}
	}
#else
	/* For testing purposes */
	got_interface = TRUE;
	if_get_config_parse_line(
"Link encap:Ethernet  HWaddr 00:08:02:c3:79:94  \n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"inet addr:69.81.112.71  Bcast:255.255.255.255  Mask:255.255.255.0\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"inet6 addr: fe80::260:8ff:fe05:e54e/64 Scope:Link\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"UP BROADCAST RUNNING MULTICAST  MTU:576  Metric:1\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"RX packets:10429960 errors:123 dropped:456 overruns:789 frame:10\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"TX packets:1275664 errors:611 dropped:311 overruns:119 carrier:911\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"collisions:99 txqueuelen:1000 \n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"RX bytes:1349414452 (1.2 GiB)  TX bytes:250404636 (238.8 MiB)\n",
		if_name,
		config_buf
	);
	if_get_config_parse_line(
"Interrupt:16 Base address:0x10AE \n",
		if_name,
		config_buf
	);
#endif

	if(got_interface)
	{
		return(0);
	}
	else
	{
		errno = ENODEV;
		return(-1);
	}
}


/*
 *	Gets the interface's state.
 *
 *	The if_name specifies the interface.
 *
 *	Returns TRUE if the interface is up or FALSE with errno set
 *	to ENODEV if the interface is down. Returns FALSE on all other
 *	errors with errno set appropriately.
 */
gboolean if_get_state(const gchar *if_name)
{
	FILE *fp;
	gboolean status;
	gint error_code;
	const gchar	*ifconfig_path = PROG_IFCONFIG;
	gchar           *cmd,
			buf[1024];

	if(if_name == NULL)
	{
		errno = EINVAL;
		return(FALSE);
	}

	/* Format the command to list all interfaces that are up */
	cmd = g_strdup_printf(
		"\"%s\" 2> \"%s\"",
		ifconfig_path,
		_PATH_DEVNULL
	);

	/* Execute the list interfaces command */
	fp = popen((const char *)cmd, "r");
	error_code = (gint)errno;

	g_free(cmd);

	if(fp == NULL)
	{
		errno = (int)error_code;
		return(FALSE);
	}

	/* Check if the specified interface was listed */
	status = FALSE;
	while(fgets(buf, sizeof(buf), fp) != NULL)
	{
		if(STRPFX(buf, if_name))
			status = TRUE;
	}

	/* Close the stream and wait for the process to exit */
	(void)pclose(fp);

	if(!status)
		errno = ENODEV;

	return(status);
}


/*
 *	Gets the interface's configuration.
 *
 *	The if_name specifies the interface.
 *
 *	The config_buf specifies the return value for the interface's
 *	configuration.
 *
 *	Returns 0 on success or non-zero on error. If the interface is
 *	down or not configured then -1 is returned with errno set to
 *	ENODEV.
 */
gint if_get_config(
	const gchar *if_name,
	IFConfig *config_buf
)
{
	FILE *fp;
	gint		status,
			error_code;
	const gchar *ifconfig_path = PROG_IFCONFIG;
	gchar *cmd;

	if((if_name == NULL) || (config_buf == NULL))
	{
		errno = EINVAL;
		return(-2);
	}

	/* Reset the return values */
	(void)memset(
		config_buf,
		0x00,
		sizeof(IFConfig)
	);

	/* Format the get interface configurations command */
	cmd = g_strdup_printf(
		"\"%s\" \"%s\" 2> \"%s\"",
		ifconfig_path,
		if_name,
		_PATH_DEVNULL
	);

	/* Execute the get interface configurations command and
	 * open the standard output stream for reading
	 */
	fp = popen((const char *)cmd, "r");
	error_code = (gint)errno;

	g_free(cmd);

	if(fp == NULL)
	{
		errno = (int)error_code;
		return(-1);
	}

	/* Get the interface's values from the stream */
	status = if_get_config_stream(
		fp,
		if_name,
		config_buf
	);
	error_code = (gint)errno;

	/* Close the stream and wait for the process to exit */
	(void)pclose(fp);

	errno = (int)error_code;

	return(status);
}
