#include <time.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "if_config.h"
#include "if_con_info.h"
#include "if_stat.h"
#include "route.h"
#include "ifm_win_pixmap.h"
#include "ifm_win_graph.h"
#include "ifm_prop_dlg.h"
#include "config.h"
#include "ifm_tooltip_messages.h"

#include "images/icon_modem_20x20.xpm"
#include "images/icon_network_settings_32x32.xpm"
#include "images/icon_hardware_settings_32x32.xpm"
#include "images/icon_close_20x20.xpm"

#include "images/icon_ifmon_48x48.xpm"


/*
 *	Flags:
 */
typedef enum {
	IFM_PROP_DLG_MAPPED             = (1 << 0),
	IFM_PROP_DLG_REALIZED           = (1 << 1)
} IFMPropDlgFlags;


/* Callbacks */
static void ifm_prop_dlg_data_destroy_cb(gpointer data);
static void ifm_prop_dlg_realize_cb(GtkWidget *widget, gpointer data);
static gint ifm_prop_dlg_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint ifm_prop_dlg_time_update_timeout_cb(gpointer data);
static void ifm_prop_dlg_route_table_select_row_cb(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
);
static void ifm_prop_dlg_route_table_unselect_row_cb(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
);
static void ifm_prop_dlg_close_cb(GtkWidget *widget, gpointer data);

/* Operations */
static void ifm_prop_dlg_update_duration(
	IFMPropDlg *d,
	const gulong cur_sec,
	const gulong con_time_start_sec
);
static void ifm_prop_dlg_route_set_values(
	IFMPropDlg *d,
	const RouteItem *item
);

/* IFMPropDlg */
static void ifm_prop_dlg_create_general_page(
	IFMPropDlg *d,
	GtkWidget *parent
);
static void ifm_prop_dlg_create_statistics_page(
	IFMPropDlg *d,
	GtkWidget *parent
);
static void ifm_prop_dlg_create_configuration_page(
	IFMPropDlg *d,
	GtkWidget *parent
);
static void ifm_prop_dlg_create_route_page(
	IFMPropDlg *d,
	GtkWidget *parent
);
IFMPropDlg *ifm_prop_dlg_new(
	const GdkRectangle *geometry
);
GtkWidget *ifm_prop_dlg_get_toplevel(IFMPropDlg *d);
void ifm_prop_dlg_set_interface_name(
	IFMPropDlg *d,
	const gchar *if_name
);
void ifm_prop_dlg_set_stats(
	IFMPropDlg *d,
	IFStat *stat_buf
);
void ifm_prop_dlg_record_history(
	IFMPropDlg *d,
	const gdouble elapsed_sec,
	const gfloat tx_coeff,
	const gfloat rx_coeff
);
void ifm_prop_dlg_set_state(
	IFMPropDlg *d,
	const gboolean state
);
void ifm_prop_dlg_set_interface_configuration(
	IFMPropDlg *d,
	const IFConfig *config
);
void ifm_prop_dlg_set_con_info(
	IFMPropDlg *d,
	const gchar *chat_report_path,
	const IFConInfo *con_info,
	const gulong bandwidth_bytes_per_sec,
	const gboolean user_set_bandwidth
);
void ifm_prop_dlg_route_table_append(
	IFMPropDlg *d,
	const RouteItem *item
);
void ifm_prop_dlg_route_table_clear(IFMPropDlg *d);
void ifm_prop_dlg_set_wm_icon(
	IFMPropDlg *d,
	IFMWinPixmap *p
);
void ifm_prop_dlg_set_state_icon(
	IFMPropDlg *d,
	IFMWinPixmap *p
);
void ifm_prop_dlg_update_display(IFMPropDlg *d);
gboolean ifm_prop_dlg_is_mapped(IFMPropDlg *d);
void ifm_prop_dlg_map(IFMPropDlg *d);
void ifm_prop_dlg_unmap(IFMPropDlg *d);
void ifm_prop_dlg_delete(IFMPropDlg *d);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define IFM_PROP_DLG_KEY		"IFMPropDlg"

#define IFM_PROP_DLG_WIDTH		380
#define IFM_PROP_DLG_HEIGHT		450


/*
 *	Properties Dialog:
 */
struct _IFMPropDlg {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		busy_count,
			freeze_count;
	IFMPropDlgFlags	flags;

	IFMWinPixmap	*last_wm_icon;

	GtkWidget	*notebook,		/* Main GtkNotebook */
			*close_btn;

	gchar		*if_name;
	gboolean	if_state;
	gulong		con_start_time_sec;	/* Connectioned time in
						 * seconds since EPOCH */

	guint		time_update_timeout_id;

	/* General Page */
	IFMWinPixmap	*last_if_state_icon;
	GtkWidget	*if_name_label,
			*if_state_icon_pixmap,		/* GtkPixmap */
			*con_duration_toplevel,
			*con_duration_entry,
			*if_address_entry,
			*con_toplevel,
			*con_speed_entry,
			*con_flags_entry,
			*if_tx_bytes_entry,
			*if_rx_bytes_entry,
			*if_tx_packets_entry,
			*if_rx_packets_entry,
			*if_tx_dropped_entry,
			*if_rx_dropped_entry,
			*if_tx_errors_entry,
			*if_rx_errors_entry,
			*if_tx_carrier_entry,
			*if_rx_frame_entry;
	IFMWinGraph	*traffic_tx_history_graph,
			*traffic_rx_history_graph;

	/* Statistics Page */
	GtkWidget	*if_stat_tx_bytes_entry,
			*if_stat_rx_bytes_entry,
			*if_stat_tx_packets_entry,
			*if_stat_rx_packets_entry,
			*if_stat_tx_compressed_entry,
			*if_stat_rx_compressed_entry,
			*if_stat_rx_multicast_entry,
			*if_stat_tx_dropped_entry,
			*if_stat_rx_dropped_entry,
			*if_stat_tx_fifo_entry,
			*if_stat_rx_fifo_entry,
			*if_stat_tx_colls_entry,
			*if_stat_tx_errors_entry,
			*if_stat_rx_errors_entry,
			*if_stat_tx_carrier_entry,
			*if_stat_rx_frame_entry;


	/* Configuration Page */
	GtkWidget	*if_config_link_encap_entry,
			*if_config_flags_entry,
			*if_config_inet_addr_entry,
			*if_config_bcast_entry,
			*if_config_ptp_entry,
			*if_config_mask_entry,
			*if_config_mtu_entry,
			*if_config_metric_entry,
			*if_config_tx_queue_len_entry,
			*if_config_interrupt_entry,
			*if_config_base_address_entry,
			*if_config_hw_addr_entry;

	/* Route Page */
	GtkWidget	*route_table_clist,
			*route_refcnt_entry,
			*route_use_entry,
			*route_metric_entry,
			*route_mtu_entry,
			*route_window_entry,
			*route_irtt_entry;
};



/*
 *	Toplevel GtkWindow data "destroy" signal callback.
 */
static void ifm_prop_dlg_data_destroy_cb(gpointer data)
{
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if(d == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_data_destroy_cb(): d=%p\n", d);
#endif

	/* Handle this signal even if already frozen */

	d->freeze_count++;

	d->time_update_timeout_id = GTK_TIMEOUT_REMOVE(d->time_update_timeout_id);

	/* The Traffic History Graphs are automatically deleted when
	 * their toplevel GtkWidgets are destroyed due to its parents
	 * being destroyed
	 */
/*	d->traffic_tx_history_graph = NULL; */
/*	d->traffic_rx_history_graph = NULL; */

	/* Unref the GtkAccelGroup */
	GTK_ACCEL_GROUP_UNREF(d->accelgrp);

	/* Delete the rest of the IFMPropDlg */
	g_free(d->if_name);

	d->freeze_count--;

	g_free(d);
}

/*
 *	Toplevel GtkWindow "realize" signal callback.
 */
static void ifm_prop_dlg_realize_cb(GtkWidget *widget, gpointer data)
{
	const gint border_major = 5;
	GdkGeometry geo;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*toplevel,
			*parent, *parent2;
	GtkNotebook *notebook;
	IFMPropDlg *d = IFM_PROP_DLG(data);
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_realize_cb(): d=%p\n", d);
#endif
	if((widget == NULL) || (d == NULL))
		return;

	window = widget->window;
	if((window == NULL) || (d->flags & IFM_PROP_DLG_REALIZED))
		return;

	/* Handle this signal even if already frozen */

	d->freeze_count++;

	accelgrp = d->accelgrp;
	toplevel = parent = d->toplevel;

	/* Set up the toplevel GtkWindow's GdkWindow */
	geo.min_width = 10;
	geo.min_height = 10;
	geo.base_width = 0;
	geo.base_height = 0;
	geo.width_inc = 1;
	geo.height_inc = 1;
	gdk_window_set_geometry_hints(
		window,
		&geo,
		GDK_HINT_MIN_SIZE |
		GDK_HINT_BASE_SIZE |
		GDK_HINT_RESIZE_INC
	);
	gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER |
		GDK_DECOR_TITLE |
		GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	);
	gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE |
		GDK_FUNC_MINIMIZE |
		GDK_FUNC_CLOSE
	);
	GUISetWMIcon(
		window,
		(guint8 **)icon_ifmon_48x48_xpm
	);

	/* Main GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Main notebook GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Main GtkNotebook */
	d->notebook = w = gtk_notebook_new();
	notebook = GTK_NOTEBOOK(w);
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(w), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), TRUE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(w), TRUE);
	gtk_widget_show(w);
	parent2 = w;

	/* General Page */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, IFM_PROP_DLG_GENERAL_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		notebook,
		w,
		gtk_label_new("General")
	);
	gtk_widget_show(w);
	ifm_prop_dlg_create_general_page(
		d,
		w
	);

	/* Statistics Page */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, IFM_PROP_DLG_STATISTICS_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		notebook,
		w,
		gtk_label_new("Statistics")
	);
	gtk_widget_show(w);
	ifm_prop_dlg_create_statistics_page(
		d,
		w
	);

	/* Configuration Page */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, IFM_PROP_DLG_CONFIGURATION_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		notebook,
		w,
		gtk_label_new("Configuration")
	);
	gtk_widget_show(w);
	ifm_prop_dlg_create_configuration_page(
		d,
		w
	);

	/* Route Page */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, IFM_PROP_DLG_ROUTE_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		notebook,
		w,
		gtk_label_new("Route")
	);
	gtk_widget_show(w);
	ifm_prop_dlg_create_route_page(
		d,
		w
	);



	/* GtkSeparator */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Buttons GtkHBox */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;


	/* Close GtkButton */
	d->close_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_close_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
"Cierre"
#elif defined(PROG_LANGUAGE_FRENCH)
"Quitter"
#elif defined(PROG_LANGUAGE_GERMAN)
"Nah"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Vicino"
#elif defined(PROG_LANGUAGE_DUTCH)
"Einde"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Prximo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Nr"
#else
"Close"
#endif
		,
		NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
		w,
		GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, border_major);
	gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(ifm_prop_dlg_close_cb), d
	);
	gtk_accel_group_add(
		accelgrp,
		GDK_Escape, 0,
		GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
		accelgrp,
		GDK_c, GDK_CONTROL_MASK,
		GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
		accelgrp,
		GDK_Return, 0,
		GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);

	/* Mark the IFMPropDlg as realized */
	d->flags |= IFM_PROP_DLG_REALIZED;

	/* Set the time update timeout callback for updating time
	 * values
	 */
	d->time_update_timeout_id = gtk_timeout_add(
		1000l,
		ifm_prop_dlg_time_update_timeout_cb, d
	);

	/* Set the initial title and values on the GtkWidgets */
	ifm_prop_dlg_update_display(d);

	d->freeze_count--;
}

/*
 *	Any GtkWidget any event signal callback.
 */
static gint ifm_prop_dlg_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
		return(status);

	switch((gint)event->type)
	{
	  case GDK_DELETE:
		if(widget == d->toplevel)
		{
			ifm_prop_dlg_unmap(d);
			status = TRUE;
		}
		break;
	}
	return(status);
}

/*
 *	Time update timeout callback.
 */
static gint ifm_prop_dlg_time_update_timeout_cb(gpointer data)
{
	gulong cur_sec;
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if(d == NULL)
		return(FALSE);

	if(d->freeze_count > 0)
		return(TRUE);

	d->freeze_count++;

	/* Get the current time */
	cur_sec = (gulong)time(NULL);

	/* Update the displayed duration */
	ifm_prop_dlg_update_duration(
		d,
		cur_sec,
		d->con_start_time_sec
	);

	d->freeze_count--;

	return(TRUE);
}

/*
 *	Route Table GtkCList "select_row" signal callback.
 */
static void ifm_prop_dlg_route_table_select_row_cb(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
)
{
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if((clist == NULL) || (d == NULL))
		return;

	if(d->freeze_count > 0)
		return;

	ifm_prop_dlg_route_set_values(
		d,
		ROUTE_ITEM(gtk_clist_get_row_data(
			clist,
			row
		))
	);
}

/*
 *	Route Table GtkCList "select_row" signal callback.
 */
static void ifm_prop_dlg_route_table_unselect_row_cb(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
)
{
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if((clist == NULL) || (d == NULL))
		return;

	if(d->freeze_count > 0)
		return;


}

/*
 *	Close callback.
 */
static void ifm_prop_dlg_close_cb(GtkWidget *widget, gpointer data)
{
	IFMPropDlg *d = IFM_PROP_DLG(data);
	if(d == NULL)
		return;

	if(d->freeze_count > 0)
		return;

	ifm_prop_dlg_unmap(d);
}


/*
 *	Updates the displayed connection duration.
 */
static void ifm_prop_dlg_update_duration(
	IFMPropDlg *d,
	const gulong cur_sec,
	const gulong con_time_start_sec
)
{
	gchar *s;
	GtkWidget *w = d->con_duration_entry;
	if(w == NULL)
		return;

	d->freeze_count++;

	if((con_time_start_sec != 0l) &&
	   (cur_sec >= con_time_start_sec)
	)
	{
		const gulong	duration_sec = cur_sec - con_time_start_sec,
				days = duration_sec / (24l * 60l * 60l),
				hours = (duration_sec / (60l * 60l)) % 24l,
				minutes = (duration_sec / 60l) % 60l,
				seconds = duration_sec % 60l;
		if(days > 0l)
			s = g_strdup_printf(
				"%ld %s %.2ld:%.2ld:%.2ld",
				days,
				(days == 1l) ? "day" : "days",
				hours, minutes, seconds
			);
		else
			s = g_strdup_printf(
				"%.2ld:%.2ld:%.2ld",
				hours, minutes, seconds
			);
	}
	else
	{
		s = g_strdup("00:00:00");
	}
	if(s != NULL)
	{
		gtk_entry_set_text(
			GTK_ENTRY(w),
			s
		);
		g_free(s);
	}

	d->freeze_count--;
}

/*
 *	Sets the route item values.
 */
static void ifm_prop_dlg_route_set_values(
	IFMPropDlg *d,
	const RouteItem *item
)
{
	d->freeze_count++;

	if(item != NULL)
	{

#define SET_ENTRY_L(_entry_,_l_)	{	\
 if((_entry_) != NULL) {			\
  gchar *s = g_strdup_printf("%ld", (_l_));	\
  gtk_entry_set_text(GTK_ENTRY(_entry_), s);	\
  g_free(s);					\
 }						\
}
		SET_ENTRY_L(
			d->route_refcnt_entry,
			item->refcnt
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_use_entry),
			(item->use) ? "Yes" : "No"
		);
		SET_ENTRY_L(
			d->route_metric_entry,
			item->metric
		);
		SET_ENTRY_L(
			d->route_mtu_entry,
			item->mtu
		);
		SET_ENTRY_L(
			d->route_window_entry,
			item->window
		);
		SET_ENTRY_L(
			d->route_irtt_entry,
			item->irtt
		);
#undef SET_ENTRY_L
	}
	else
	{
		gtk_entry_set_text(
			GTK_ENTRY(d->route_refcnt_entry),
			""
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_use_entry),
			""
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_metric_entry),
			""
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_mtu_entry),
			""
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_window_entry),
			""
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->route_irtt_entry),
			""
		);
	}

	d->freeze_count--;
}


/*
 *	Creates the general page.
 */
static void ifm_prop_dlg_create_general_page(
	IFMPropDlg *d,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
			border_minor = 2;
	gint font_size;
	const gchar *font_encoding;
	gchar *font_name_h1_bold;
	GdkFont *font;
	GdkPixmap *pixmap;
	GtkRcStyle *rcstyle;
	GtkWidget	*w,
			*parent2, *parent3, *parent4, *parent5,
			*toplevel = d->toplevel;
	GdkWindow *window = toplevel->window;
	GtkStyle *style = gtk_widget_get_style(toplevel);
	IFMWinGraph *graph;

	/* Get the base font size */
	font = style->font;
	font_size = GDK_FONT_GET_FONT_NAME_SIZE(font);
	if(font_size < 3)
	    font_size = 3;

	/* Format the font names */
#if defined(PROG_LANGUAGE_POLISH)
	font_encoding = "iso8859-2";
#else
	font_encoding = "iso8859-1";
#endif

	font_name_h1_bold = g_strdup_printf(
"-adobe-helvetica-bold-r-normal-*-%i-*-*-*-*-*-%s",
	    font_size + 2,
	    font_encoding
	);


#define NEW_ENTRY(_parent_,_entry_,_label_,_tip_,_lwidth_,_ewidth_,_units_label_) {	\
 GtkWidget *_parent2, *_parent3;				\
								\
 w = gtk_hbox_new(FALSE, border_minor);				\
 if(GTK_IS_VBOX(_parent_))					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else if((_ewidth_) > 0)					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else								\
  gtk_box_pack_start(GTK_BOX(_parent_), w, TRUE, TRUE, 0);	\
 gtk_widget_show(w);						\
 _parent2 = w;							\
								\
 /* GtkLabel */							\
 if((_lwidth_) > 0) {						\
  w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);		\
  gtk_widget_set_usize(w, (_lwidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
  _parent3 = w;							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_container_add(GTK_CONTAINER(_parent3), w);		\
  gtk_widget_show(w);						\
 } else {							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
								\
 /* GtkEntry */							\
 (_entry_) = w = gtk_entry_new();				\
 gtk_entry_set_editable(GTK_ENTRY(w), FALSE);			\
 if((_ewidth_) > 0) {						\
  gtk_widget_set_usize(w, (_ewidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
 } else {							\
  gtk_box_pack_start(GTK_BOX(_parent2), w, TRUE, TRUE, 0);	\
 }								\
 if((_tip_) != NULL)						\
  GUISetWidgetTip(w, (_tip_));					\
 gtk_widget_show(w);						\
								\
 /* Units GtkLabel */						\
 if((_units_label_) != NULL) {					\
  w = gtk_label_new(_units_label_);				\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
}

	w = gtk_frame_new("Interface");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	/* State Icon & Label GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* State Icon & Label GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* State Icon GtkPixmap */
	pixmap = gdk_pixmap_new(window, 32, 32, -1);
	d->if_state_icon_pixmap = w = gtk_pixmap_new(
		pixmap, NULL
	);
	(void)GDK_PIXMAP_UNREF(pixmap);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* State Icon GtkLabel */
	d->if_name_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	rcstyle = gtk_rc_style_new();
	rcstyle->font_name = STRDUP(font_name_h1_bold);
	gtk_widget_modify_style(w, rcstyle);
	GTK_RC_STYLE_UNREF(rcstyle);
	gtk_widget_show(w);

	/* Duration GtkEntry */
	NEW_ENTRY(
		parent4,
		d->con_duration_entry,
		"Duration:",
		IFM_TTM_CON_DURATION,
		55,
		120,
		NULL
	);
	d->con_duration_toplevel = d->con_duration_entry->parent;

	/* Interface Address GtkEntry */
	NEW_ENTRY(
		parent4,
		d->if_address_entry,
		"Address:",
		IFM_TTM_INET_ADDR,
		55,
		120,
		NULL
	);


	/* Traffic History Graph GtkVBox */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_end(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Traffic History TX IFMWinGraph */
	d->traffic_tx_history_graph = graph = ifm_win_graph_new(
		parent4,
		GTK_DIR_LEFT,
		IFM_PROP_DLG_TX_TRAFFIC_HISTORY_GRAPH_WIDGET_NAME,
		"TX",
		140, 40,
		60				/* History time in seconds */
	);

	/* Traffic History RX IFMWinGraph */
	d->traffic_rx_history_graph = graph = ifm_win_graph_new(
		parent4,
		GTK_DIR_LEFT,
		IFM_PROP_DLG_RX_TRAFFIC_HISTORY_GRAPH_WIDGET_NAME,
		"RX",
		140, 40,
		60				/* History time in seconds */
	);


	/* Connection GtkFrame */
	d->con_toplevel = w = gtk_frame_new("Connection");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->con_speed_entry,
		"Speed:",
		IFM_TTM_CON_SPEED_KBPS,
		-1,
		70,
		"kbps"
	);
	NEW_ENTRY(
		parent3,
		d->con_flags_entry,
		"Flags:",
		IFM_TTM_CON_FLAGS,
		-1,
		-1,
		NULL
	);


	/* Bytes & Packets GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_frame_new("Bytes");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_tx_bytes_entry,
		"TX:",
		IFM_TTM_TX_BYTES,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_rx_bytes_entry,
		"RX:",
		IFM_TTM_RX_BYTES,
		30,
		120,
		NULL
	);


	w = gtk_frame_new("Packets");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_tx_packets_entry,
		"TX:",
		IFM_TTM_TX_PACKETS,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_rx_packets_entry,
		"RX:",
		IFM_TTM_RX_PACKETS,
		30,
		120,
		NULL
	);


	/* Dropped & Errors GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Dropped GtkFrame */
	w = gtk_frame_new("Dropped");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_tx_dropped_entry,
		"TX:",
		IFM_TTM_TX_DROPPED,
		30,
		60,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_rx_dropped_entry,
		"RX:",
		IFM_TTM_RX_DROPPED,
		30,
		60,
		NULL
	);

	/* Errors GtkFrame */
	w = gtk_frame_new("Errors");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	NEW_ENTRY(
		parent4,
		d->if_tx_errors_entry,
		"TX:",
		IFM_TTM_TX_ERRORS,
		30,
		60,
		NULL
	);
	NEW_ENTRY(
		parent4,
		d->if_rx_errors_entry,
		"RX:",
		IFM_TTM_RX_ERRORS,
		30,
		60,
		NULL
	);

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	NEW_ENTRY(
		parent4,
		d->if_tx_carrier_entry,
		"Carrier:",
		IFM_TTM_TX_CARRIER,
		50,
		-1,
		NULL
	);
	NEW_ENTRY(
		parent4,
		d->if_rx_frame_entry,
		"Frame:",
		IFM_TTM_RX_FRAME,
		50,
		-1,
		NULL
	);

#undef NEW_ENTRY

	g_free(font_name_h1_bold);
}

/*
 *	Creates the Statistics Page.
 */
static void ifm_prop_dlg_create_statistics_page(
	IFMPropDlg *d,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
			border_minor = 2;
	GtkWidget	*w,
			*parent2, *parent3, *parent4;

#define NEW_ENTRY(_parent_,_entry_,_label_,_tip_,_lwidth_,_ewidth_,_units_label_) {	\
 GtkWidget *_parent2, *_parent3;				\
								\
 w = gtk_hbox_new(FALSE, border_minor);				\
 if(GTK_IS_VBOX(_parent_))					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else if((_ewidth_) > 0)					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else								\
  gtk_box_pack_start(GTK_BOX(_parent_), w, TRUE, TRUE, 0);	\
 gtk_widget_show(w);						\
 _parent2 = w;							\
								\
 /* GtkLabel */							\
 if((_lwidth_) > 0) {						\
  w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);		\
  gtk_widget_set_usize(w, (_lwidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
  _parent3 = w;							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_container_add(GTK_CONTAINER(_parent3), w);		\
  gtk_widget_show(w);						\
 } else {							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
								\
 /* GtkEntry */							\
 (_entry_) = w = gtk_entry_new();				\
 gtk_entry_set_editable(GTK_ENTRY(w), FALSE);			\
 if((_ewidth_) > 0) {						\
  gtk_widget_set_usize(w, (_ewidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
 } else {							\
  gtk_box_pack_start(GTK_BOX(_parent2), w, TRUE, TRUE, 0);	\
 }								\
 if((_tip_) != NULL)						\
  GUISetWidgetTip(w, (_tip_));					\
 gtk_widget_show(w);						\
								\
 /* Units GtkLabel */						\
 if((_units_label_) != NULL) {					\
  w = gtk_label_new(_units_label_);				\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
}

	/* Bytes & Packets GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_frame_new("Bytes");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_bytes_entry,
		"TX:",
		IFM_TTM_TX_BYTES,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_stat_rx_bytes_entry,
		"RX:",
		IFM_TTM_RX_BYTES,
		30,
		120,
		NULL
	);


	w = gtk_frame_new("Packets");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_packets_entry,
		"TX:",
		IFM_TTM_TX_PACKETS,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_stat_rx_packets_entry,
		"RX:",
		IFM_TTM_RX_PACKETS,
		30,
		120,
		NULL
	);


	/* Compressed & Multicast GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_frame_new("Compressed");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_compressed_entry,
		"TX:",
		IFM_TTM_TX_COMPRESSED,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_stat_rx_compressed_entry,
		"RX:",
		IFM_TTM_RX_COMPRESSED,
		30,
		120,
		NULL
	);


	w = gtk_frame_new("Multicast");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* No TX for Multicast */

	NEW_ENTRY(
		parent3,
		d->if_stat_rx_multicast_entry,
		"RX:",
		IFM_TTM_RX_MULTICAST,
		30,
		120,
		NULL
	);


	/* Dropped, FIFO, Collisions  GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Dropped GtkFrame */
	w = gtk_frame_new("Dropped");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_dropped_entry,
		"TX:",
		IFM_TTM_TX_DROPPED,
		30,
		-1,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_stat_rx_dropped_entry,
		"RX:",
		IFM_TTM_RX_DROPPED,
		30,
		-1,
		NULL
	);

	/* FIFO Errors GtkFrame */
	w = gtk_frame_new("FIFO Errors");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_fifo_entry,
		"TX:",
		IFM_TTM_TX_FIFO,
		30,
		-1,
		NULL
	);
	NEW_ENTRY(
		parent3,
		d->if_stat_rx_fifo_entry,
		"RX:",
		IFM_TTM_RX_FIFO,
		30,
		-1,
		NULL
	);

	/* Collisions GtkFrame */
	w = gtk_frame_new("Collisions");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_stat_tx_colls_entry,
		"TX:",
		IFM_TTM_TX_COLLISIONS,
		30,
		-1,
		NULL
	);

	/* No RX collisions */


	/* Errors GtkHbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Errors GtkFrame */
	w = gtk_frame_new("Errors");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	NEW_ENTRY(
		parent4,
		d->if_stat_tx_errors_entry,
		"TX:",
		IFM_TTM_TX_ERRORS,
		30,
		120,
		NULL
	);
	NEW_ENTRY(
		parent4,
		d->if_stat_rx_errors_entry,
		"RX:",
		IFM_TTM_RX_ERRORS,
		30,
		120,
		NULL
	);

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	NEW_ENTRY(
		parent4,
		d->if_stat_tx_carrier_entry,
		"Carrier:",
		IFM_TTM_TX_CARRIER,
		50,
		-1,
		NULL
	);
	NEW_ENTRY(
		parent4,
		d->if_stat_rx_frame_entry,
		"Frame:",
		IFM_TTM_RX_FRAME,
		50,
		-1,
		NULL
	);

#undef NEW_ENTRY
}

/*
 *	Creates the configuration page.
 */
static void ifm_prop_dlg_create_configuration_page(
	IFMPropDlg *d,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
			border_minor = 2;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkWidget	*w,
			*parent2, *parent3, *parent4,
			*toplevel = d->toplevel;
	GdkWindow	*window = toplevel->window;

#define NEW_ENTRY(_parent_,_entry_,_label_,_tip_,_lwidth_,_ewidth_,_units_label_) {	\
 GtkWidget *_parent2, *_parent3;				\
								\
 w = gtk_hbox_new(FALSE, border_minor);				\
 if(GTK_IS_VBOX(_parent_))					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else if((_ewidth_) > 0)					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else								\
  gtk_box_pack_start(GTK_BOX(_parent_), w, TRUE, TRUE, 0);	\
 gtk_widget_show(w);						\
 _parent2 = w;							\
								\
 /* GtkLabel */							\
 if((_lwidth_) > 0) {						\
  w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);		\
  gtk_widget_set_usize(w, (_lwidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
  _parent3 = w;							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_container_add(GTK_CONTAINER(_parent3), w);		\
  gtk_widget_show(w);						\
 } else {							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
								\
 /* GtkEntry */							\
 (_entry_) = w = gtk_entry_new();				\
 gtk_entry_set_editable(GTK_ENTRY(w), FALSE);			\
 if((_ewidth_) > 0) {						\
  gtk_widget_set_usize(w, (_ewidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
 } else {							\
  gtk_box_pack_start(GTK_BOX(_parent2), w, TRUE, TRUE, 0);	\
 }								\
 if((_tip_) != NULL)						\
  GUISetWidgetTip(w, (_tip_));					\
 gtk_widget_show(w);						\
								\
 /* Units GtkLabel */						\
 if((_units_label_) != NULL) {					\
  w = gtk_label_new(_units_label_);				\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
}

	/* Type GtkFrame */
	w = gtk_frame_new("Type");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Type Icon GtkPixmap */
	pixmap = gdk_pixmap_create_from_xpm_d(
		window,
		&mask,
		NULL,
		(gchar **)icon_modem_20x20_xpm
	);
	w = gtk_pixmap_new(
		pixmap, mask
	);
	(void)GDK_PIXMAP_UNREF(pixmap);
	(void)GDK_BITMAP_UNREF(mask);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Link Encap GtkEntry */
	d->if_config_link_encap_entry = w = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	GUISetWidgetTip(
		w,
		IFM_TTM_LINK_ENCAP
	);
	gtk_widget_show(w);


	/* Flags GtkFrame */
	w = gtk_frame_new("Flags");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	d->if_config_flags_entry = w = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Addresses GtkFrame */
	w = gtk_frame_new("Addresses");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_inet_addr_entry,
		"Internet:",
		IFM_TTM_INET_ADDR,
		50,
		105,
		NULL
	);

	NEW_ENTRY(
		parent3,
		d->if_config_bcast_entry,
		"Broadcast:",
		IFM_TTM_BCAST,
		60,
		105,
		NULL
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_ptp_entry,
		"PTP:",
		IFM_TTM_PTP,
		50,
		105,
		NULL
	);

	NEW_ENTRY(
		parent3,
		d->if_config_mask_entry,
		"Mask:",
		IFM_TTM_MASK,
		60,
		105,
		NULL
	);

	/* Network Settings GtkFrame */
	w = gtk_frame_new("Network Settings");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Icon GtkPixmap */
	pixmap = gdk_pixmap_create_from_xpm_d(
		window,
		&mask,
		NULL,
		(gchar **)icon_network_settings_32x32_xpm
	);
	w = gtk_pixmap_new(
		pixmap, mask
	);
	(void)GDK_PIXMAP_UNREF(pixmap);
	(void)GDK_BITMAP_UNREF(mask);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_mtu_entry,
		"MTU:",
		IFM_TTM_MTU,
		-1,
		80,
		"bytes"
	);

	NEW_ENTRY(
		parent3,
		d->if_config_metric_entry,
		"Metric:",
		IFM_TTM_METRIC,
		-1,
		60,
		NULL
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_tx_queue_len_entry,
		"TX Queue Len:",
		IFM_TTM_TX_QUEUE_LEN,
		-1,
		80,
		"bytes"
	);


	/* Hardware Settings GtkFrame */
	w = gtk_frame_new("Hardware Settings");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Icon GtkPixmap */
	pixmap = gdk_pixmap_create_from_xpm_d(
		window,
		&mask,
		NULL,
		(gchar **)icon_hardware_settings_32x32_xpm
	);
	w = gtk_pixmap_new(
		pixmap, mask
	);
	(void)GDK_PIXMAP_UNREF(pixmap);
	(void)GDK_BITMAP_UNREF(mask);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_interrupt_entry,
		"IRQ:",
		IFM_TTM_INTERRUPT,
		-1,
		60,
		NULL
	);

	NEW_ENTRY(
		parent3,
		d->if_config_base_address_entry,
		"Base Address:",
		IFM_TTM_BASE_ADDRESS,
		-1,
		90,
		NULL
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ENTRY(
		parent3,
		d->if_config_hw_addr_entry,
		"Serial Number:",
		IFM_TTM_HW_ADDR,
		-1,
		120,
		NULL
	);


#undef NEW_ENTRY

}

/*
 *	Creates the route page.
 */
static void ifm_prop_dlg_create_route_page(
	IFMPropDlg *d,
	GtkWidget *parent
)
{
	const gint	border_minor = 2,
			border_major = 5;
	gchar *heading[5];
	GtkWidget	*w,
			*parent2, *parent3;
	GtkCList *clist;

#define NEW_ENTRY(_parent_,_entry_,_label_,_lwidth_,_ewidth_,_units_label_) {	\
 GtkWidget *_parent2, *_parent3;				\
								\
 w = gtk_hbox_new(FALSE, border_minor);				\
 if(GTK_IS_VBOX(_parent_))					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else if((_ewidth_) > 0)					\
  gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0);	\
 else								\
  gtk_box_pack_start(GTK_BOX(_parent_), w, TRUE, TRUE, 0);	\
 gtk_widget_show(w);						\
 _parent2 = w;							\
								\
 /* GtkLabel */							\
 if((_lwidth_) > 0) {						\
  w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);		\
  gtk_widget_set_usize(w, (_lwidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
  _parent3 = w;							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_container_add(GTK_CONTAINER(_parent3), w);		\
  gtk_widget_show(w);						\
 } else {							\
  w = gtk_label_new(_label_);					\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
								\
 /* GtkEntry */							\
 (_entry_) = w = gtk_entry_new();				\
 gtk_entry_set_editable(GTK_ENTRY(w), FALSE);			\
 if((_ewidth_) > 0) {						\
  gtk_widget_set_usize(w, (_ewidth_), -1);			\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
 } else {							\
  gtk_box_pack_start(GTK_BOX(_parent2), w, TRUE, TRUE, 0);	\
 }								\
 gtk_widget_show(w);						\
								\
 /* Units GtkLabel */						\
 if((_units_label_) != NULL) {					\
  w = gtk_label_new(_units_label_);				\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);	\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);	\
  gtk_widget_show(w);						\
 }								\
}

	/* Route Table GtkFrame */
	w = gtk_frame_new("Route Table");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkScrolledWindow for GtkCList */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Route Table GtkCList */
	heading[0] = "Interface";
	heading[1] = "Destination";
	heading[2] = "Gateway";
	heading[3] = "Mask";
	heading[4] = "Flags";
	d->route_table_clist = w = gtk_clist_new_with_titles(
		sizeof(heading) / sizeof(gchar *),
		heading
	);
	clist = GTK_CLIST(w);
	gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(ifm_prop_dlg_route_table_select_row_cb), d
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(ifm_prop_dlg_route_table_unselect_row_cb), d
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_BROWSE);
	if(clist->columns > 0)
	{
		const gint ncolumns = clist->columns;
		gint i;
		for(i = 0; i < ncolumns; i++)
		{
		        gtk_clist_set_column_justification(
		                clist,
				i,
				GTK_JUSTIFY_LEFT
		        );
			gtk_clist_set_column_resizeable(
				clist,
				i,
				TRUE
			);
		        gtk_clist_set_column_auto_resize(
				clist,
				i,
				TRUE
			);
		}
	}
	gtk_clist_set_row_height(clist, 20);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_widget_show(w);


	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Reference Count */
	NEW_ENTRY(
		parent3,
		d->route_refcnt_entry,
		"Ref Count:",
		-1,
		60,
		NULL
	);

	/* Useable */
	NEW_ENTRY(
		parent3,
		d->route_use_entry,
		"Useable:",
		-1,
		30,
		NULL
	);

	/* Metric */
	NEW_ENTRY(
		parent3,
		d->route_metric_entry,
		"Metric:",
		-1,
		60,
		NULL
	);


	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* MTU */
	NEW_ENTRY(
		parent3,
		d->route_mtu_entry,
		"MTU:",
		-1,
		60,
		NULL
	);

	/* Window */
	NEW_ENTRY(
		parent3,
		d->route_window_entry,
		"Window:",
		-1,
		60,
		NULL
	);

	/* IRTT */
	NEW_ENTRY(
		parent3,
		d->route_irtt_entry,
		"IRTT:",
		-1,
		60,
		"ms"
	);

#undef NEW_ENTRY
}

/*
 *	Creates a new IFMPropDlg.
 */
IFMPropDlg *ifm_prop_dlg_new(
	const GdkRectangle *geometry
)
{
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*toplevel;
	IFMPropDlg *d = IFM_PROP_DLG(g_malloc0(sizeof(IFMPropDlg)));
	if(d == NULL)
		return(NULL);

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_new(): d=%p\n", d);
#endif

	d->accelgrp = accelgrp = gtk_accel_group_new();
	d->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#if 0
	d->busy_count = 0;
	d->freeze_count = 0;
#endif

	d->freeze_count++;

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_object_set_data_full(
		GTK_OBJECT(w), IFM_PROP_DLG_KEY,
		d, ifm_prop_dlg_data_destroy_cb
	);
	gtk_window_set_wmclass(
		GTK_WINDOW(w),
		IFM_PROP_DLG_WM_WINDOW_NAME,	/* Window name */
		IFM_WM_CLASS_NAME		/* Class name */
	);
	gtk_widget_set_name(w, IFM_PROP_DLG_TOPLEVEL_WIDGET_NAME);
	gtk_window_set_policy(GTK_WINDOW(w), FALSE, FALSE, FALSE);
	gtk_widget_set_usize(
		w,
		IFM_PROP_DLG_WIDTH, IFM_PROP_DLG_HEIGHT
	);
	if(geometry != NULL)
		gtk_widget_set_uposition(
			w,
			(gint)geometry->x, (gint)geometry->y
		);
	gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_FOCUS_CHANGE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "realize",
		GTK_SIGNAL_FUNC(ifm_prop_dlg_realize_cb), d
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "delete_event",
		GTK_SIGNAL_FUNC(ifm_prop_dlg_event_cb), d
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);

	/* Call ifm_prop_dlg_realize_cb to create the rest of the
	 * IFMPropDlg
	 */
	gtk_widget_realize(w);

	d->freeze_count--;

	return(d);
}

/*
 *	Sets the toplevel GtkWindow.
 */
GtkWidget *ifm_prop_dlg_get_toplevel(IFMPropDlg *d)
{
	if(d == NULL)
		return(NULL);

	return(d->toplevel);
}

/*
 *	Sets the interface name,
 */
void ifm_prop_dlg_set_interface_name(
	IFMPropDlg *d,
	const gchar *if_name
)
{
	if((d == NULL) || (if_name == NULL))
		return;

	if(d->if_name == if_name)
		return;

	d->freeze_count++;

	g_free(d->if_name);
	d->if_name = g_strdup(if_name);

	/* General Page */
	if(d->if_name_label != NULL)
		gtk_label_set_text(
			GTK_LABEL(d->if_name_label),
			if_name
		);

	/* Update the displayed title */
	ifm_prop_dlg_update_display(d);

	d->freeze_count--;
}

/*
 *	Sets the interface statistics.
 */
void ifm_prop_dlg_set_stats(
	IFMPropDlg *d,
	IFStat *stat_buf
)
{
	if((d == NULL) || (stat_buf == NULL))
		return;

	d->freeze_count++;

#define SET_ENTRY_L(_entry_,_l_)	{	\
 if((_entry_) != NULL) {			\
  gchar *s = g_strdup_printf("%ld", (_l_));	\
  gtk_entry_set_text(GTK_ENTRY(_entry_), s);	\
  g_free(s);					\
 }						\
}
	/* General Page */
	SET_ENTRY_L(d->if_tx_bytes_entry, stat_buf->tx_bytes);
	SET_ENTRY_L(d->if_rx_bytes_entry, stat_buf->rx_bytes);

	SET_ENTRY_L(d->if_tx_packets_entry, stat_buf->tx_packets);
	SET_ENTRY_L(d->if_rx_packets_entry, stat_buf->rx_packets);

	SET_ENTRY_L(d->if_tx_dropped_entry, stat_buf->tx_dropped);
	SET_ENTRY_L(d->if_rx_dropped_entry, stat_buf->rx_dropped);

	SET_ENTRY_L(d->if_tx_errors_entry, stat_buf->tx_errors);
	SET_ENTRY_L(d->if_rx_errors_entry, stat_buf->rx_errors);

	SET_ENTRY_L(d->if_tx_carrier_entry, stat_buf->tx_carrier);
	SET_ENTRY_L(d->if_rx_frame_entry, stat_buf->rx_frame);

	/* Statistics Page */
	SET_ENTRY_L(d->if_stat_tx_bytes_entry, stat_buf->tx_bytes);
	SET_ENTRY_L(d->if_stat_rx_bytes_entry, stat_buf->rx_bytes);

	SET_ENTRY_L(d->if_stat_tx_packets_entry, stat_buf->tx_packets);
	SET_ENTRY_L(d->if_stat_rx_packets_entry, stat_buf->rx_packets);

	SET_ENTRY_L(d->if_stat_tx_compressed_entry, stat_buf->tx_compressed);
	SET_ENTRY_L(d->if_stat_rx_compressed_entry, stat_buf->rx_compressed);

	SET_ENTRY_L(d->if_stat_rx_multicast_entry, stat_buf->rx_multicast);

	SET_ENTRY_L(d->if_stat_tx_dropped_entry, stat_buf->tx_dropped);
	SET_ENTRY_L(d->if_stat_rx_dropped_entry, stat_buf->rx_dropped);

	SET_ENTRY_L(d->if_stat_tx_fifo_entry, stat_buf->tx_fifo);
	SET_ENTRY_L(d->if_stat_rx_fifo_entry, stat_buf->rx_fifo);

	SET_ENTRY_L(d->if_stat_tx_colls_entry, stat_buf->tx_colls);

	SET_ENTRY_L(d->if_stat_tx_errors_entry, stat_buf->tx_errors);
	SET_ENTRY_L(d->if_stat_rx_errors_entry, stat_buf->rx_errors);

	SET_ENTRY_L(d->if_stat_tx_carrier_entry, stat_buf->tx_carrier);
	SET_ENTRY_L(d->if_stat_rx_frame_entry, stat_buf->rx_frame);

#undef SET_ENTRY_L

	d->freeze_count--;
}

/*
 *	Records the traffic history.
 */
void ifm_prop_dlg_record_history(
	IFMPropDlg *d,
	const gdouble elapsed_sec,
	const gfloat tx_coeff,
	const gfloat rx_coeff
)
{
	if(d == NULL)
		return;

	d->freeze_count++;

	/* General Page */
	ifm_win_graph_record(
		d->traffic_tx_history_graph,
		elapsed_sec,
		tx_coeff
	);
	ifm_win_graph_record(
		d->traffic_rx_history_graph,
		elapsed_sec,
		rx_coeff
	);

	d->freeze_count--;
}

/*
 *	Sets the interface state.
 */
void ifm_prop_dlg_set_state(
	IFMPropDlg *d,
	const gboolean state
)
{
	if(d == NULL)
		return;

	d->freeze_count++;

	d->if_state = state;

	d->freeze_count--;
}

/*
 *	Sets the interface's configuration.
 */
void ifm_prop_dlg_set_interface_configuration(
	IFMPropDlg *d,
	const IFConfig *config
)
{
	if((d == NULL) || (config == NULL))
		return;

	d->freeze_count++;

	/* General Page */
	if(d->if_address_entry != NULL)
	{
		const INAddress *inet_addr = &config->inet_addr;
		gchar *s = g_strdup_printf(
			"%i.%i.%i.%i",
			inet_addr->u8[0],
			inet_addr->u8[1],
			inet_addr->u8[2],
			inet_addr->u8[3]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_address_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	/* Configuration Page */
	if(d->if_config_link_encap_entry != NULL)
	{
		const gchar *s = config->link_encap;
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_link_encap_entry),
			(s != NULL) ? s : ""
		);
	}

	if(d->if_config_hw_addr_entry != NULL)
	{
		const IFHardwareAddress *hw_addr = &config->hw_addr;
		gchar *s = g_strdup_printf(
			"%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
			hw_addr->u8[0],
			hw_addr->u8[1],
			hw_addr->u8[2],
			hw_addr->u8[3],
			hw_addr->u8[4],
			hw_addr->u8[5]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_hw_addr_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_flags_entry != NULL)
	{
		const IFConfigFlags flags = config->flags;
		gchar *s = g_strconcat(
			(flags & IF_CONFIG_FLAG_UP) ? "UP " : "",
			(flags & IF_CONFIG_FLAG_LOOPBACK) ? "LOOPBACK " : "",
			(flags & IF_CONFIG_FLAG_BROADCAST) ? "BROADCAST " : "",
			(flags & IF_CONFIG_FLAG_POINTTOPOINT) ? "POINTTOPOINT " : "",
			(flags & IF_CONFIG_FLAG_NOTRAILERS) ? "NOTRAILERS " : "",
			(flags & IF_CONFIG_FLAG_RUNNING) ? "RUNNING " : "",
			(flags & IF_CONFIG_FLAG_NOARP) ? "NOARP " : "",
			(flags & IF_CONFIG_FLAG_MULTICAST) ? "MULTICAST " : "",
			NULL
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_flags_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_inet_addr_entry != NULL)
	{
		const INAddress *addr = &config->inet_addr;
		gchar *s = g_strdup_printf(
			"%d.%d.%d.%d",
			addr->u8[0],
			addr->u8[1],
			addr->u8[2],
			addr->u8[3]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_inet_addr_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_bcast_entry != NULL)
	{
		const INAddress *addr = &config->bcast;
		gchar *s = g_strdup_printf(
			"%d.%d.%d.%d",
			addr->u8[0],
			addr->u8[1],
			addr->u8[2],
			addr->u8[3]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_bcast_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_ptp_entry != NULL)
	{
		const INAddress *addr = &config->ptp;
		gchar *s = g_strdup_printf(
			"%d.%d.%d.%d",
			addr->u8[0],
			addr->u8[1],
			addr->u8[2],
			addr->u8[3]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_ptp_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_mask_entry != NULL)
	{
		const INAddress *addr = &config->mask;
		gchar *s = g_strdup_printf(
			"%d.%d.%d.%d",
			addr->u8[0],
			addr->u8[1],
			addr->u8[2],
			addr->u8[3]
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_mask_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_mtu_entry != NULL)
	{
		gchar *s = g_strdup_printf(
			"%ld",
			config->mtu
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_mtu_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_metric_entry != NULL)
	{
		gchar *s = g_strdup_printf(
			"%ld",
			config->metric
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_metric_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_tx_queue_len_entry != NULL)
	{
		gchar *s = g_strdup_printf(
			"%ld",
			config->tx_queue_len
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_tx_queue_len_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_interrupt_entry != NULL)
	{
		gchar *s = g_strdup_printf(
			"%i",
			config->interrupt
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_interrupt_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	if(d->if_config_base_address_entry != NULL)
	{
		const INAddress *addr = &config->base_address;
		gchar *s = g_strdup_printf(
			"0x%.8X",
			addr->u32
		);
		gtk_entry_set_text(
			GTK_ENTRY(d->if_config_base_address_entry),
			(s != NULL) ? s : ""
		);
		g_free(s);
	}

	d->freeze_count--;
}

/*
 *	Sets the connection information.
 *
 *	If chat_report_path is a non-empty string then it specifies
 *	the path to the connection information and that the values
 *	in con_info are known and will be used to display the connection
 *	information. Otherwise the value of bandwidth_bytes_per_sec will
 *	be used to display the connection information.
 *
 *	The con_info specifies the connection information values.
 *
 *	The bandwidth_bytes_per_sec specifies the bandwidth in bytes
 *	per second. If chat_report_path is not a non-empty string then
 *	bandwidth_bytes_per_sec will be used to display the connection
 *	information.
 *
 *	If user_set_bandwidth is TRUE then it suggests that the user
 *	has explicitly set the bandwidth and all bandwidth values will
 *	be displayed with a note indicating that the user has
 *	explicitly set it.
 */
void ifm_prop_dlg_set_con_info(
	IFMPropDlg *d,
	const gchar *chat_report_path,
	const IFConInfo *con_info,
	const gulong bandwidth_bytes_per_sec,
	const gboolean user_set_bandwidth
)
{
	if((d == NULL) || (con_info == NULL))
		return;

	d->freeze_count++;

#define SET_ENTRY_L(_entry_,_l_)	{	\
 if((_entry_) != NULL) {			\
  gchar *s = g_strdup_printf("%ld", (_l_));	\
  gtk_entry_set_text(GTK_ENTRY(_entry_), s);	\
  g_free(s);					\
 }						\
}

	/* Connection start time */
	d->con_start_time_sec = con_info->start_time_sec;

	/* General Page */

	/* Set the duration sensitive if the connection start time
	 * is known
	 */
	GTK_WIDGET_SET_SENSITIVE(
		d->con_duration_toplevel,
		(d->con_start_time_sec != 0l) ? TRUE : FALSE
	);


	/* Bandwidth and connection speed
	 *
	 * Is the source of the connection information not specified?
	 */
	if(STRISEMPTY(chat_report_path))
	{
		/* Use the bandwidth as the connection speed */
		SET_ENTRY_L(
			d->con_speed_entry,
			bandwidth_bytes_per_sec * 8l
		);
		GTK_WIDGET_SET_SENSITIVE(
			d->con_toplevel,
			(bandwidth_bytes_per_sec != 0l) ? TRUE : FALSE
		);
	}
	else
	{
		SET_ENTRY_L(
			d->con_speed_entry,
			con_info->speed_bits_per_sec
		);
		GTK_WIDGET_SET_SENSITIVE(
			d->con_toplevel,
			(con_info->speed_bits_per_sec != 0l) ? TRUE : FALSE   
		);
	}

	/* Connection flags */
	gtk_entry_set_text(
		GTK_ENTRY(d->con_flags_entry),
		con_info->flags_string
	);

#undef SET_ENTRY_L

	d->freeze_count--;
}

/*
 *	Appends an item to the route table.
 */
void ifm_prop_dlg_route_table_append(
	IFMPropDlg *d,
	const RouteItem *item
)
{
	gint		i,
			column,
			row,
			ncolumns;
	gchar **strv;
	GtkCList *clist;

	if((d == NULL) || (item == NULL))
		return;

	d->freeze_count++;

	clist = GTK_CLIST(d->route_table_clist);
	ncolumns = MAX(clist->columns, 1);

	gtk_clist_freeze(clist);

	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	for(i = 0; i < ncolumns; i++)
		strv[i] = "";

	row = gtk_clist_append(
		clist,
		strv
	);
	if(row > -1)
	{
		/* Interface */
		column = 0;
		if(column < ncolumns)
		{
			gtk_clist_set_text(
				clist,
				row, column,
				item->iface
			);
		}

		/* Destination */
		column = 1;
		if(column < ncolumns)
		{
			const INAddress *addr = &item->destination;
			gchar *s = g_strdup_printf(
"%d.%d.%d.%d",
				addr->u8[0],
				addr->u8[1],
				addr->u8[2],
				addr->u8[3]
			);
			gtk_clist_set_text(
				clist,
				row, column,
				s
			);
			g_free(s);
		}

		/* Gateway */
		column = 2;
		if(column < ncolumns)
		{
			const INAddress *addr = &item->gateway;
			gchar *s = g_strdup_printf(
"%d.%d.%d.%d",
				addr->u8[0],
				addr->u8[1],
				addr->u8[2],
				addr->u8[3]
			);
			gtk_clist_set_text(
				clist,
				row, column,
				s
			);
			g_free(s);
		}

		/* Mask */
		column = 3;
		if(column < ncolumns)
		{
			const INAddress *addr = &item->mask;
			gchar *s = g_strdup_printf(
"%d.%d.%d.%d",
				addr->u8[0],
				addr->u8[1],
				addr->u8[2],
				addr->u8[3]
			);
			gtk_clist_set_text(
				clist,
				row, column,
				s
			);
			g_free(s);
		}

		/* Flags */
		column = 4;
		if(column < ncolumns)
		{
			const RouteFlags flags = item->flags;
			gchar *s = g_strconcat(
			        (flags & ROUTE_FLAG_UP) ? "U" : "",
		                (flags & ROUTE_FLAG_GATEWAY) ? "G" : "",
	                        (flags & ROUTE_FLAG_HOST) ? "H" : "",
				(flags & ROUTE_FLAG_REINSTATE) ? "R" : "",
			        (flags & ROUTE_FLAG_DYNAMIC) ? "D" : "",
		                (flags & ROUTE_FLAG_MODIFIED) ? "M" : "",
	                        (flags & ROUTE_FLAG_MTU) ? "T" : "",
				(flags & ROUTE_FLAG_WINDOW) ? "W" : "",
			        (flags & ROUTE_FLAG_IRTT) ? "I" : "",
		                (flags & ROUTE_FLAG_REJECT) ? "!" : "",
				NULL
			);
			gtk_clist_set_text(
				clist,
				row, column,
				s
			);
			g_free(s);
		}

		/* Metric */
		column = 5;
		if(column < ncolumns)
		{
			gchar *s = g_strdup_printf(
				"%ld",
				item->metric
			);
			gtk_clist_set_text(
				clist,
				row, column,
				s
			);
			g_free(s);
		}

		/* Set the RouteItem as the row data
		 *
		 * It is same to use g_memdup() in the RouteItem
		 * since it has no pointers to dynamically allocated
		 * members
		 */
		gtk_clist_set_row_data_full(
			clist,
			row,
			g_memdup(
				item,
				sizeof(RouteItem)
			),
			g_free
		);

		/* Automatically display the values for the first row
		 * appended
		 */
		if(row == 0)
			ifm_prop_dlg_route_set_values(
				d,
				item
			);
	}

	g_free(strv);

	gtk_clist_columns_autosize(clist);
	gtk_clist_thaw(clist);

	d->freeze_count--;
}

/*
 *	Clears the route table.
 */
void ifm_prop_dlg_route_table_clear(IFMPropDlg *d)
{
	GtkCList *clist;

	if(d == NULL)
		return;

	d->freeze_count++;

	clist = GTK_CLIST(d->route_table_clist);

	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_columns_autosize(clist);
	gtk_clist_thaw(clist);

	ifm_prop_dlg_route_set_values(
		d,
		NULL
	);

	d->freeze_count--;
}

/*
 *	Sets the WM icon.
 */
void ifm_prop_dlg_set_wm_icon(
	IFMPropDlg *d,
	IFMWinPixmap *p
)
{
	GdkWindow *window;
	GtkWidget *w;

	if((d == NULL) || (p == NULL))
		return;

	if(d->last_wm_icon == p)
		return;

	if(!ifm_win_pixmap_is_loaded(p))
		return;

	w = d->toplevel;
	window = w->window;
	if(window == NULL)
		return;

	d->freeze_count++;

	gdk_window_set_icon(
		window,
		window,
		p->pixmap, p->mask
	);
	d->last_wm_icon = p;
	gdk_flush();

	d->freeze_count--;
}

/*
 *	Sets the state icon.
 */
void ifm_prop_dlg_set_state_icon(
	IFMPropDlg *d,
	IFMWinPixmap *p
)
{
	GtkWidget *w;

	if((d == NULL) || (p == NULL))
		return;

	if(d->last_if_state_icon == p)
		return;

	if(!ifm_win_pixmap_is_loaded(p))
		return;

	w = d->if_state_icon_pixmap;
	if(w == NULL)
		return;

	d->freeze_count++;

	gtk_pixmap_set(
		GTK_PIXMAP(w),
		p->pixmap, p->mask
	);
	d->last_if_state_icon = p;

	d->freeze_count--;
}


/*
 *	Updates the displayed values on the widgets.
 */
void ifm_prop_dlg_update_display(IFMPropDlg *d)
{
	gboolean if_state;
	gulong con_start_time_sec;
	gchar *s;

	if(d == NULL)
		return;

	d->freeze_count++;

	if_state = d->if_state;
	con_start_time_sec = d->con_start_time_sec;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_update_display(): d=%p\n", d);
#endif

	/* Update the title */
	if(d->if_name != NULL)
		s = g_strdup_printf(
			"Interface Details: %s",
			d->if_name
		);
	else
		s = g_strdup("Interface Details");
	gtk_window_set_title(
		GTK_WINDOW(d->toplevel),
		(s != NULL) ? s : ""
	);
	g_free(s);



	d->freeze_count--;
}

/*
 *	Checks if the IFMPropDlg is mapped.
 */
gboolean ifm_prop_dlg_is_mapped(IFMPropDlg *d)
{
	if(d == NULL)
		return(FALSE);

	return((d->flags & IFM_PROP_DLG_MAPPED) ? TRUE : FALSE);
}

/*
 *	Maps the IFMPropDlg.
 */
void ifm_prop_dlg_map(IFMPropDlg *d)
{
	if(d == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_map() %p\n", d);
#endif

	gtk_widget_show(d->toplevel);
	d->flags |= IFM_PROP_DLG_MAPPED;
}

/*
 *	Unmaps the IFMPropDlg.
 */
void ifm_prop_dlg_unmap(IFMPropDlg *d)
{
	if(d == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_unmap() %p\n", d);
#endif
	gtk_widget_hide(d->toplevel);
	d->flags &= ~IFM_PROP_DLG_MAPPED;
}

/*
 *	Deletes the IFMPropDlg.
 */
void ifm_prop_dlg_delete(IFMPropDlg *d)
{
	if(d == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_prop_dlg_delete() %p\n", d);
#endif

	ifm_prop_dlg_unmap(d);

	/* Destroy the toplevel GtkWindow and call ifm_prop_dlg_data_destroy_cb()
	 * to delete the rest of the IFMPropDlg
	 */
	GTK_WIDGET_DESTROY(d->toplevel);
}
