#include <errno.h>
#include <time.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "process.h"
#include "in_address.h"
#include "if_config.h"
#include "if_con_info.h"
#include "if_stat.h"
#include "route.h"
#include "guiutils.h"
#include "statictip.h"
#include "menu_button.h"
#include "cdialog.h"
#include "ifm_win_pixmap.h"
#include "ifm_win_gauge.h"
#include "ifm_win_graph.h"
#include "ifm_win.h"
#include "config.h"

#include "images/icon_close_20x20.xpm"

#include "images/icon_if_state_down_20x20.xpm"
#include "images/icon_if_state_connecting_20x20.xpm"
#include "images/icon_if_state_up_20x20.xpm"
#include "images/icon_if_state_up_rx_20x20.xpm"
#include "images/icon_if_state_up_tx_20x20.xpm"
#include "images/icon_if_state_up_rx_tx_20x20.xpm"

#include "images/icon_if_state_down_32x32.xpm"
#include "images/icon_if_state_connecting_32x32.xpm"
#include "images/icon_if_state_up_32x32.xpm"
#include "images/icon_if_state_up_rx_32x32.xpm"
#include "images/icon_if_state_up_tx_32x32.xpm"
#include "images/icon_if_state_up_rx_tx_32x32.xpm"

#include "images/icon_if_state_down_48x48.xpm"
#include "images/icon_if_state_connecting_48x48.xpm"
#include "images/icon_if_state_up_48x48.xpm"
#include "images/icon_if_state_up_rx_48x48.xpm"
#include "images/icon_if_state_up_tx_48x48.xpm"
#include "images/icon_if_state_up_rx_tx_48x48.xpm"

#include "images/icon_ifmon_48x48.xpm"

#include "images/icon_connect_20x20.xpm"
#include "images/icon_disconnect_20x20.xpm"
#include "images/icon_reconnect_20x20.xpm"
#include "images/icon_details_20x20.xpm"


/* Callbacks */
static void ifm_win_data_destroy_cb(gpointer data);
static void ifm_win_realize_cb(GtkWidget *widget, gpointer data);
static gint ifm_win_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint ifm_win_interface_state_btn_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint ifm_win_traffic_history_graph_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint ifm_win_update_stats_timeout_cb(gpointer data);
static gint ifm_win_update_state_timeout_cb(gpointer data);
static void ifm_win_connect_cb(GtkWidget *widget, gpointer data);
static void ifm_win_disconnect_cb(GtkWidget *widget, gpointer data);
static void ifm_win_reconnect_cb(GtkWidget *widget, gpointer data);
static void ifm_win_details_cb(GtkWidget *widget, gpointer data);
static void ifm_win_show_interface_state_toggle_cb(GtkWidget *widget, gpointer data);
static void ifm_win_show_traffic_history_toggle_cb(GtkWidget *widget, gpointer data);
static void ifm_win_show_traffic_gauges_toggle_cb(GtkWidget *widget, gpointer data);
static void ifm_win_gauge_style_continuous_cb(GtkWidget *widget, gpointer data) ;
static void ifm_win_gauge_style_leds_cb(GtkWidget *widget, gpointer data);
static void ifm_win_gauge_style_gradient_leds_cb(GtkWidget *widget, gpointer data);
static void ifm_win_close_cb(GtkWidget *widget, gpointer data);

/* Operations */
static void ifm_win_connect(IFMWin *win);
static void ifm_win_disconnect(IFMWin *win);
static void ifm_win_reconnect(IFMWin *win);
static void ifm_win_details(IFMWin *win);

/* IFMWin */
IFMWin *ifm_win_new(
	const gchar *if_name,
	const gulong bandwidth_bytes_per_sec,
	const gboolean user_set_bandwidth,
	const GtkOrientation orientation,
	const gboolean show_border,
	const gboolean show_title,
	const gboolean animate_wm_icon,
	const IFMWinDisplay display,
	const IFMWinGaugeStyle gauge_style,
	const gint gauge_length,
	const gint gauge_nticks,
	const gulong stats_update_int_ms,
	const gchar *chat_report_path
);
void ifm_win_set_display(
	IFMWin *win,
	const IFMWinDisplay display
);
void ifm_win_set_gauge_style(
	IFMWin *win,
	const IFMWinGaugeStyle gauge_style
);
void ifm_win_set_connect_command(
	IFMWin *win,
	const gchar *cmd
);
void ifm_win_set_disconnect_command(
	IFMWin *win,
	const gchar *cmd
);
void ifm_win_set_reconnect_command(
	IFMWin *win,
	const gchar *cmd
);
void ifm_win_update_stats(IFMWin *win);
void ifm_win_update_state(IFMWin *win);
IFMWinStateIconCodes ifm_win_get_state_icon_code(IFMWin *win);
void ifm_win_set_state_icons_by_code(
	IFMWin *win,
	const IFMWinStateIconCodes state_icon_code
);
static gchar *ifm_win_format_duration_string(const gulong duration_sec);
static gchar *ifm_win_format_tip_string(IFMWin *win);
static void ifm_win_update_interface_state_tip(IFMWin *win);
static void ifm_win_update_traffic_history_tip(IFMWin *win);
void ifm_win_update_display(IFMWin *win);
gboolean ifm_win_is_mapped(IFMWin *win);
void ifm_win_map(IFMWin *win);
void ifm_win_unmap(IFMWin *win);
void ifm_win_delete(IFMWin *win);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define IFM_WIN_KEY			"IFMWin"


/*
 *	Toplevel GtkWindow data "destroy" signal callback.
 */
static void ifm_win_data_destroy_cb(gpointer data)
{
	gint i;
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_data_destroy_cb(): win=%p\n", win);
#endif

	/* Handle this signal even if already frozen */

	win->freeze_count++;

	/* Remove the timeout callbacks */
	win->stats_update_timeout_id = GTK_TIMEOUT_REMOVE(win->stats_update_timeout_id);
	win->state_update_timeout_id = GTK_TIMEOUT_REMOVE(win->state_update_timeout_id);

	/* Delete the Details IFMPropDlg */
	ifm_prop_dlg_delete(win->prop_dlg);
	win->prop_dlg = NULL;

	/* The Interface State GtkMenu was destroyed when the Interface
	 * State Menu Button was destroyed
	 */
/*	win->state_menu = NULL; */

	/* Destroy the Right Click GtkMenu */
	GTK_WIDGET_DESTROY(win->menu);

	/* The Traffic History IFMWinGraph is automatically deleted
	 * when their toplevel GtkWidgets are destroyed due to its
	 * parents being destroyed
	 */
/*	win->traffic_history_graph = NULL; */

	/* IFMWinGauges will delete themselves when their destroy
	 * callbacks are called
	 */
/*	win->tx_traffic_gauge = NULL; */
/*	win->rx_traffic_gauge = NULL; */

	/* Unref the GtkAccelGroup */
	GTK_ACCEL_GROUP_UNREF(win->accelgrp);

	/* Destroy the GdkCursors */
	gdk_cursor_destroy(win->translate_cur);

	/* Unref all the icons */
	for(i = 0; i < IFM_WIN_NSTATE_ICONS; i++)
	{
		win->wm_icon_pixmaps[i] = ifm_win_pixmap_unref(win->wm_icon_pixmaps[i]);
		win->state_icon_pixmaps[i] = ifm_win_pixmap_unref(win->state_icon_pixmaps[i]);
		win->prop_dlg_state_icon_pixmaps[i] = ifm_win_pixmap_unref(win->prop_dlg_state_icon_pixmaps[i]);
	}
/*	win->last_wm_icon_pixmap = NULL; */
/*	win->last_state_icon_pixmap = NULL; */

	/* Delete the rest of the IFMWin */
	g_free(win->if_name);
	g_free(win->chat_report_path);
	g_timer_destroy(win->stat_timer);
	g_free(win->connect_cmd);
	g_free(win->disconnect_cmd);
	g_free(win->reconnect_cmd);
	g_free(win->last_prop_dlg_geometry);

	win->freeze_count--;

	g_free(win);
}

/*
 *	Toplevel GtkWindow "realize" signal callback.
 */
static void ifm_win_realize_cb(GtkWidget *widget, gpointer data)
{
	const gint	border_major = 5,
			border_minor = 2;
	gboolean vertical;
	GdkGeometry geo;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*toplevel,
			*menu,
			*parent, *parent2, *parent3;
	IFMWinPixmap *p;
	IFMWinGraph *graph;
	IFMWin *win = IFM_WIN(data);
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_realize_cb(): win=%p\n", win);
#endif
	if((widget == NULL) || (win == NULL))
		return;

	window = widget->window;
	if((window == NULL) || (win->flags & IFM_WIN_REALIZED))
		return;

	/* Handle this signal even if already frozen */

	win->freeze_count++;

	vertical = (win->orientation == GTK_ORIENTATION_VERTICAL) ?
		TRUE : FALSE;
	accelgrp = win->accelgrp;
	toplevel = parent = win->toplevel;

	/* Set up the toplevel GtkWindow's GdkWindow */
	geo.min_width = 10;
	geo.min_height = 10;
	geo.base_width = 0;
	geo.base_height = 0;
	geo.width_inc = 1;
	geo.height_inc = 1;
	gdk_window_set_geometry_hints(
		window,
		&geo,
		GDK_HINT_MIN_SIZE |
		GDK_HINT_BASE_SIZE |
		GDK_HINT_RESIZE_INC
	);
	gdk_window_set_decorations(
		window,
		((win->flags & IFM_WIN_SHOW_BORDER) ?  GDK_DECOR_BORDER : 0) |
		((win->flags & IFM_WIN_SHOW_TITLE) ?  GDK_DECOR_TITLE : 0)
	);
	gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_CLOSE
	);
#define LOAD_ICON(_icons_list_,_state_,_xpm_data_) {	\
 IFMWinPixmap *p = ifm_win_pixmap_new_from_xpm_data(	\
  window,						\
  (guint8 **)(_xpm_data_)				\
 );							\
 (_icons_list_)[_state_] = p;				\
}
	if(win->flags & IFM_WIN_ANIMATE_WM_ICON)
	{
		/* Load the WM icon pixmaps */
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_DOWN,
			icon_if_state_down_48x48_xpm
		);
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_CONNECTING,
			icon_if_state_connecting_48x48_xpm
		);
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_UP,
			icon_if_state_up_48x48_xpm
		);
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_UP_RX,
			icon_if_state_up_rx_48x48_xpm
		);
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_UP_TX,
			icon_if_state_up_tx_48x48_xpm
		);
		LOAD_ICON(
			win->wm_icon_pixmaps,
			IFM_WIN_STATE_ICON_UP_RX_TX,
			icon_if_state_up_rx_tx_48x48_xpm
		);
	}

	/* Load the state icon pixmaps */
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_DOWN,
		icon_if_state_down_20x20_xpm
	);
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_CONNECTING,
		icon_if_state_connecting_20x20_xpm
	);
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP,
		icon_if_state_up_20x20_xpm
	);
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_RX,
		icon_if_state_up_rx_20x20_xpm
	);
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_TX,
		icon_if_state_up_tx_20x20_xpm
	);
	LOAD_ICON(
		win->state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_RX_TX,
		icon_if_state_up_rx_tx_20x20_xpm
	);

	/* Load the Details IFMPropDlg state icon pixmaps */
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_DOWN,
		icon_if_state_down_32x32_xpm
	);
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_CONNECTING,
		icon_if_state_connecting_32x32_xpm
	);
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP,
		icon_if_state_up_32x32_xpm
	);
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_RX,
		icon_if_state_up_rx_32x32_xpm
	);
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_TX,
		icon_if_state_up_tx_32x32_xpm
	);
	LOAD_ICON(
		win->prop_dlg_state_icon_pixmaps,
		IFM_WIN_STATE_ICON_UP_RX_TX,
		icon_if_state_up_rx_tx_32x32_xpm
	);
#undef LOAD_ICON

	if(!(win->flags & IFM_WIN_ANIMATE_WM_ICON))
		GUISetWMIcon(
			window,
			(guint8 **)icon_ifmon_48x48_xpm
		);

	/* Create the GdkCursors */
	win->translate_cur = gdk_cursor_new(GDK_FLEUR);

	/* If no borders are shown then create a main GtkFrame before
	 * the main GtkBox
	 */
	if(!(win->flags & IFM_WIN_SHOW_BORDER))
	{
		w = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
		gtk_container_add(GTK_CONTAINER(parent), w);
	        gtk_widget_show(w);
		parent = w;
	}

	/* Main GtkBox */
	w = vertical ?
		gtk_vbox_new(FALSE, border_major) :
		gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Interface State toplevel GtkBox */
	win->state_toplevel = w = vertical ?
		gtk_hbox_new(TRUE, 0) :
		gtk_vbox_new(TRUE, 0);
	gtk_widget_set_name(w, IFM_WIN_INTERFACE_STATE_TOPLEVEL_WIDGET_NAME);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(win->display & IFM_WIN_DISPLAY_INTERFACE_STATE)
		gtk_widget_show(w);
	parent2 = w;

	/* Interface State Menu GtkButton */
	win->state_btn = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
	gtk_widget_add_events(
		w,
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "enter_notify_event",
		GTK_SIGNAL_FUNC(ifm_win_interface_state_btn_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "leave_notify_event",
		GTK_SIGNAL_FUNC(ifm_win_interface_state_btn_event_cb), win
	);
	menu_button_new_from_button(
		w,
		&menu
	);
	menu_button_set_map_trigger(
		w,
		MENU_BUTTON_MAP_TYPE_PRESSED
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Interface State GtkMenu */
	if(menu != NULL)
	{
		guint accel_key, accel_mods;
		guint8 **icon;
		const gchar *label;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer data = win;

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_LABEL,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_CHECK,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU_LABEL	{	\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SUBMENU,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
 GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL, NULL,					\
  0, 0,						\
  NULL, NULL					\
 );						\
}
		icon = (guint8 **)icon_details_20x20_xpm;
		label = "Details...";
		accel_key = GDK_Return;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = ifm_win_details_cb;
		ADD_MENU_ITEM_LABEL
		win->state_menu_details_mi = w;

		ADD_MENU_SEPARATOR

		icon = (guint8 **)icon_connect_20x20_xpm;
		label = "Connect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_connect_cb;
		ADD_MENU_ITEM_LABEL
		win->state_menu_connect_mi = w;

		icon = (guint8 **)icon_disconnect_20x20_xpm;
		label = "Disconnect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_disconnect_cb;
		ADD_MENU_ITEM_LABEL
		win->state_menu_disconnect_mi = w;

		icon = (guint8 **)icon_reconnect_20x20_xpm;
		label = "Reconnect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_reconnect_cb;
		ADD_MENU_ITEM_LABEL
		win->state_menu_reconnect_mi = w;

		ADD_MENU_SEPARATOR

		icon = (guint8 **)icon_close_20x20_xpm;
		label =
#if defined(PROG_LANGUAGE_SPANISH)
"Cierre"
#elif defined(PROG_LANGUAGE_FRENCH)
"Fermer"
#elif defined(PROG_LANGUAGE_GERMAN)
"Nah"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Vicino"
#elif defined(PROG_LANGUAGE_DUTCH)
"Einde"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Prximo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Nr"
#else
"Close"
#endif
		;
		accel_key = GDK_w;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = ifm_win_close_cb;
		ADD_MENU_ITEM_LABEL
		win->state_menu_close_mi = w;

#undef ADD_MENU_SEPARATOR
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SUBMENU_LABEL
#undef ADD_MENU_ITEM_LABEL
	}

	/* Create the state icon GtkPixmap on the state GtkButton */
	p = win->state_icon_pixmaps[IFM_WIN_STATE_ICON_DOWN];
	if(ifm_win_pixmap_is_loaded(p))
	{
		GtkWidget *w;
		win->state_icon_pm = w = gtk_pixmap_new(
			p->pixmap, p->mask
		);
		gtk_container_add(GTK_CONTAINER(parent3), w);
		gtk_widget_show(w);
	}


	/* Traffic History Graph toplevel GtkBox */
	win->traffic_history_toplevel = w = vertical ?
		gtk_hbox_new(TRUE, 0) :
		gtk_vbox_new(TRUE, 0);
	gtk_widget_set_name(w, IFM_WIN_TRAFFIC_HISTORY_TOPLEVEL_WIDGET_NAME);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(win->display & IFM_WIN_DISPLAY_TRAFFIC_HISTORY)
		gtk_widget_show(w);
	parent2 = w;

	/* Traffic History IFMWinGraph */
	win->traffic_history_graph = graph = ifm_win_graph_new(
		parent2,
		GTK_DIR_LEFT,
		IFM_WIN_TRAFFIC_HISTORY_GRAPH_WIDGET_NAME,
		NULL,
		25, 25,
		30                              /* History time in seconds */
	);
	w = ifm_win_graph_get_drawing_area(graph);
	gtk_widget_add_events(
		w,
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(ifm_win_traffic_history_graph_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(ifm_win_traffic_history_graph_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "enter_notify_event",
		GTK_SIGNAL_FUNC(ifm_win_traffic_history_graph_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "leave_notify_event",
		GTK_SIGNAL_FUNC(ifm_win_traffic_history_graph_event_cb), win
	);


	/* Traffic Gauges toplevel GtkBox */
	win->traffic_gauges_toplevel = w = vertical ?
		gtk_hbox_new(TRUE, 0) :
		gtk_vbox_new(TRUE, 0);
	gtk_widget_set_name(w, IFM_WIN_TRAFFIC_GAUGES_TOPLEVEL_WIDGET_NAME);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(win->display & IFM_WIN_DISPLAY_TRAFFIC_GAUGES)
		gtk_widget_show(w);
	parent2 = w;

	w = vertical ?
		gtk_hbox_new(FALSE, border_minor) :
		gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	if(vertical)
	{
		/* Incoming Traffic Gauge */
		win->rx_traffic_gauge = ifm_win_gauge_new(
			parent3,
			IFM_WIN_RX_TRAFFIC_GAUGE_WIDGET_NAME,
			win->orientation,
			"R", -1,
			win->gauge_style,
			win->gauge_length,
			win->gauge_nticks
		);
		/* Outgoing Traffic Gauge */
		win->tx_traffic_gauge = ifm_win_gauge_new(
			parent3,
			IFM_WIN_TX_TRAFFIC_GAUGE_WIDGET_NAME,
			win->orientation,
			"T", -1,
			win->gauge_style,
			win->gauge_length,
			win->gauge_nticks
		);
	}
	else
	{
		/* Outgoing Traffic Gauge */
		win->tx_traffic_gauge = ifm_win_gauge_new(
			parent3,
			IFM_WIN_TX_TRAFFIC_GAUGE_WIDGET_NAME,
			win->orientation,
			"TX", IFM_WIN_TRAFFIC_GAUGE_LABEL_WIDTH,
			win->gauge_style,
			win->gauge_length,
			win->gauge_nticks
		);
		/* Incoming Traffic Gauge */
		win->rx_traffic_gauge = ifm_win_gauge_new(
			parent3,
			IFM_WIN_RX_TRAFFIC_GAUGE_WIDGET_NAME,
			win->orientation,
			"RX", IFM_WIN_TRAFFIC_GAUGE_LABEL_WIDTH,
			win->gauge_style,
			win->gauge_length,
			win->gauge_nticks
		);
	}

	/* Right Click GtkMenu */
	win->menu = menu = GUIMenuCreate();
	if(menu != NULL)
	{
		guint accel_key, accel_mods;
		guint8 **icon;
		const gchar *label;
		GtkWidget *submenu;
		void (*func_cb)(GtkWidget *w, gpointer);
		gpointer data = win;

#define ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_LABEL,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_CHECK	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_CHECK,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
}
#define ADD_MENU_ITEM_SUBMENU_LABEL	{	\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SUBMENU,			\
  accelgrp,					\
  icon, label,					\
  accel_key, accel_mods,			\
  func_cb, data					\
 );						\
 GUIMenuItemSetSubMenu(w, submenu);		\
}
#define ADD_MENU_SEPARATOR	{		\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL, NULL,					\
  0, 0,						\
  NULL, NULL					\
 );						\
}
		icon = (guint8 **)icon_details_20x20_xpm;
		label = "Details...";
		accel_key = GDK_Return;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = ifm_win_details_cb;
		ADD_MENU_ITEM_LABEL
		win->details_mi = w;

		ADD_MENU_SEPARATOR

		icon = (guint8 **)icon_connect_20x20_xpm;
		label = "Connect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_connect_cb;
		ADD_MENU_ITEM_LABEL
		win->connect_mi = w;

		icon = (guint8 **)icon_disconnect_20x20_xpm;
		label = "Disconnect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_disconnect_cb;
		ADD_MENU_ITEM_LABEL
		win->disconnect_mi = w;

		icon = (guint8 **)icon_reconnect_20x20_xpm;
		label = "Reconnect";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_reconnect_cb;
		ADD_MENU_ITEM_LABEL
		win->reconnect_mi = w;

		ADD_MENU_SEPARATOR

		icon = NULL;
		label = "Show Interface State";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_show_interface_state_toggle_cb;
		ADD_MENU_ITEM_CHECK
		win->show_interface_state_micheck = w;

		icon = NULL;
		label = "Show Traffic History";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_show_traffic_history_toggle_cb;
		ADD_MENU_ITEM_CHECK
		win->show_traffic_history_micheck = w;

		icon = NULL;
		label = "Show Traffic Gauges";
		accel_key = 0;
		accel_mods = 0;
		func_cb = ifm_win_show_traffic_gauges_toggle_cb;
		ADD_MENU_ITEM_CHECK
		win->show_traffic_gauges_micheck = w;

		ADD_MENU_SEPARATOR

		submenu = GUIMenuCreate();
		if(submenu != NULL)
		{
			GtkWidget *menu = submenu;

			icon = NULL;
			label = "Continuous";
			accel_key = 0;
			accel_mods = 0;
			func_cb = ifm_win_gauge_style_continuous_cb;
			ADD_MENU_ITEM_CHECK
			win->gauge_style_continuous_micheck = w;

			icon = NULL;
			label = "Leds";
			accel_key = 0;
			accel_mods = 0;
			func_cb = ifm_win_gauge_style_leds_cb;
			ADD_MENU_ITEM_CHECK
			win->gauge_style_leds_micheck = w;

			icon = NULL;
			label = "Gradient Leds";
			accel_key = 0;
			accel_mods = 0;
			func_cb = ifm_win_gauge_style_gradient_leds_cb;
			ADD_MENU_ITEM_CHECK
			win->gauge_style_gradient_leds_micheck = w;
		}
		icon = NULL;
		label = "Gauge Style";
		accel_key = 0;
		accel_mods = 0;
		func_cb = NULL;
		ADD_MENU_ITEM_SUBMENU_LABEL

		ADD_MENU_SEPARATOR

		icon = (guint8 **)icon_close_20x20_xpm;
		label =
#if defined(PROG_LANGUAGE_SPANISH)
"Cierre"
#elif defined(PROG_LANGUAGE_FRENCH)
"Fermer"
#elif defined(PROG_LANGUAGE_GERMAN)
"Nah"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Vicino"
#elif defined(PROG_LANGUAGE_DUTCH)
"Einde"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Prximo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Nr"
#else
"Close"
#endif
		;
		accel_key = GDK_w;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = ifm_win_close_cb;
		ADD_MENU_ITEM_LABEL
		win->close_mi = w;

#undef ADD_MENU_SEPARATOR
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SUBMENU_LABEL
#undef ADD_MENU_ITEM_LABEL
	}

	/* Mark the IFMWin as realized */
	win->flags |= IFM_WIN_REALIZED;

	/* Get the initial statistics and state and then update the
	 * displayed values
	 */
	ifm_win_update_stats(win);
	ifm_win_update_state(win);

	/* Set the update and display interface statistics timeout callback */
	win->stats_update_timeout_id = gtk_timeout_add(
		MAX(
			win->stats_update_int_ms,
			IFM_WIN_MIN_STATS_UPDATE_INTERVAL
		),
		ifm_win_update_stats_timeout_cb, win
	);

	/* Set the update and display interface state timeout callback */
	win->stats_update_timeout_id = gtk_timeout_add(
		IFM_WIN_DEF_STATE_UPDATE_INTERVAL,
		ifm_win_update_state_timeout_cb, win
	);

	win->freeze_count--;
}

/*
 *	Any GtkWidget any event signal callback.
 */
static gint ifm_win_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventFocus *focus;
	GdkEventButton *button;
	GdkEventMotion *motion;
	GdkEventCrossing *crossing;
	IFMWin *win = IFM_WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
		return(status);

	if(win->freeze_count > 0)
		return(status);

	win->freeze_count++;

	switch((gint)event->type)
	{
	  case GDK_DELETE:
		if(widget == win->toplevel)
		{
			ifm_win_unmap(win);
			status = TRUE;
		}
		break;

	  case GDK_CONFIGURE:
		configure = (GdkEventConfigure *)event;
		break;

	  case GDK_EXPOSE:
		expose = (GdkEventExpose *)event;
		break;

	  case GDK_FOCUS_CHANGE:
		focus = (GdkEventFocus *)event;
		if(GTK_WIDGET_CAN_FOCUS(widget))
		{
			if(!GTK_WIDGET_HAS_FOCUS(widget) && focus->in)
				GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
			else if(GTK_WIDGET_HAS_FOCUS(widget) && !focus->in)
				GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
			status = TRUE;
		}
		break;

	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		/* Grab focus on the event widget if possible */
		if(!GTK_WIDGET_HAS_FOCUS(widget) &&
		   GTK_WIDGET_CAN_FOCUS(widget)
		)
			gtk_widget_grab_focus(widget);

		switch(button->button)
		{
		    case GDK_BUTTON1:
			if(!(win->flags & IFM_WIN_SHOW_TITLE))
			{
				GdkModifierType mask;
				gdk_window_get_pointer(
					widget->window,
					&win->last_button_x, &win->last_button_y,
					&mask
				);
				win->flags |= IFM_WIN_DRAG_MOVE;
				gdk_pointer_grab(
					button->window,
					FALSE,
					GDK_BUTTON_RELEASE_MASK |
					GDK_POINTER_MOTION_MASK,
					NULL,
					win->translate_cur,
					button->time
				);
				status = TRUE;
			}
			break;

		    case GDK_BUTTON3:
			gtk_menu_popup(
				GTK_MENU(win->menu),
				NULL, NULL,
				NULL, NULL,
				button->button, button->time
			);
			status = TRUE;
			break;
		}
		break;

	  case GDK_BUTTON_RELEASE:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		    case GDK_BUTTON1:
			if(win->flags & IFM_WIN_DRAG_MOVE)
			{
				gdk_pointer_ungrab(button->time);
				win->flags &= ~IFM_WIN_DRAG_MOVE;
				status = TRUE;
			}
			break;
		    case GDK_BUTTON3:
			status = TRUE;
			break;
		}
		break;

	  case GDK_MOTION_NOTIFY:
		motion = (GdkEventMotion *)event;
		if(win->flags & IFM_WIN_DRAG_MOVE)
		{
			if(motion->is_hint)
			{
				gint x, y;
				GdkModifierType mask;
				gdk_window_get_pointer(
					motion->window,
					&x, &y,
					&mask
				);
			}
			gtk_widget_set_uposition(
				widget,
				(gint)motion->x_root - win->last_button_x,
				(gint)motion->y_root - win->last_button_y
			);
			status = TRUE;
		}
		break;

	  case GDK_ENTER_NOTIFY:
		crossing = (GdkEventCrossing *)event;
		break;

	  case GDK_LEAVE_NOTIFY:
		break;
	}

	win->freeze_count--;

	return(status);
}

/*
 *	Interface state GtkButton any event signal callback.
 */
static gint ifm_win_interface_state_btn_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventCrossing *crossing;
	IFMWin *win = IFM_WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
		return(status);

	if(win->freeze_count > 0)
		return(status);

	win->freeze_count++;

	switch((gint)event->type)
	{
	  case GDK_ENTER_NOTIFY:
		crossing = (GdkEventCrossing *)event;
		if(!(win->flags & IFM_WIN_SHOW_INTERFACE_STATE_TIP))
			win->flags |= IFM_WIN_SHOW_INTERFACE_STATE_TIP;
		ifm_win_update_interface_state_tip(win);
		break;

	  case GDK_LEAVE_NOTIFY:
		crossing = (GdkEventCrossing *)event;
		win->flags &= ~IFM_WIN_SHOW_INTERFACE_STATE_TIP;
		StaticTipSet(
			widget,
			NULL,
			STATIC_TIP_ALIGN_POINTER_MIN, STATIC_TIP_ALIGN_POINTER_MIN,
			0, 0
		);
		break;
	}

	win->freeze_count--;

	return(status);
}

/*
 *	History graph GtkDrawingArea any event signal callback.
 */
static gint ifm_win_traffic_history_graph_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventButton *button;
	GdkEventCrossing *crossing;
	IFMWin *win = IFM_WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
		return(status);

	if(win->freeze_count > 0)
		return(status);

	win->freeze_count++;

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		    case GDK_BUTTON1:
			ifm_win_details(win);
			status = TRUE;
			break;
		}
		break;

	  case GDK_BUTTON_RELEASE:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		    case GDK_BUTTON1:
			status = TRUE;
			break;
		}
		break;

	  case GDK_ENTER_NOTIFY:
		crossing = (GdkEventCrossing *)event;
		if(!(win->flags & IFM_WIN_SHOW_TRAFFIC_HISTORY_TIP))
			win->flags |= IFM_WIN_SHOW_TRAFFIC_HISTORY_TIP;
		ifm_win_update_traffic_history_tip(win);
		break;

	  case GDK_LEAVE_NOTIFY:
		crossing = (GdkEventCrossing *)event;
		win->flags &= ~IFM_WIN_SHOW_TRAFFIC_HISTORY_TIP;
		StaticTipSet(
			widget,
			NULL,
			STATIC_TIP_ALIGN_POINTER_MIN, STATIC_TIP_ALIGN_POINTER_MIN,
			0, 0
		);
		break;
	}

	win->freeze_count--;

	return(status);
}

/*
 *	Update and display interface statistics timeout callback.
 */
static gint ifm_win_update_stats_timeout_cb(gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return(FALSE);

	if(win->freeze_count > 0)
		return(TRUE);

	win->freeze_count++;

	/* Get the current interface statistics and redraw the display
	 * as needed to reflect any changes in values
	 */
	ifm_win_update_stats(win);

	win->freeze_count--;

	return(TRUE);
}

/*
 *	Update and display interface state timeout callback.
 */
static gint ifm_win_update_state_timeout_cb(gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return(FALSE);

	if(win->freeze_count > 0)
		return(TRUE);

	win->freeze_count++;

	/* If the Details IFMPropDlg is unmapped then delete it */
	if(win->prop_dlg != NULL)
	{
		IFMPropDlg *d = win->prop_dlg;
		if(!ifm_prop_dlg_is_mapped(d))
		{
			/* Record the geometry */
			GtkWidget *w = ifm_prop_dlg_get_toplevel(d);
			if(w != NULL)
			{
		                gint x = 0, y = 0;
				GdkRectangle *geometry = win->last_prop_dlg_geometry;
				GdkWindow *window = w->window;
		                gdk_window_get_root_origin(
					window,
					&x, &y
				);
				if(geometry == NULL)
					win->last_prop_dlg_geometry = geometry = (GdkRectangle *)g_malloc0(sizeof(GdkRectangle));
				if(geometry != NULL)
				{
					geometry->x = (gint16)x;
					geometry->y = (gint16)y;
				}
			}

			/* Delete the Details IFMPropDlg */
			ifm_prop_dlg_delete(win->prop_dlg);
			win->prop_dlg = d = NULL;
		}
	}

	/* Check for any changes in the interface's state and update
	 * the displayed values
	 */
	ifm_win_update_state(win);

	win->freeze_count--;

	return(TRUE);
}

/*
 *	Connect callback.
 */
static void ifm_win_connect_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_connect(win);
}

/*
 *	Disconnect callback.
 */
static void ifm_win_disconnect_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_disconnect(win);
}

/*
 *	Reconnect callback.
 */
static void ifm_win_reconnect_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_reconnect(win);
}

/*
 *	Details callback.
 */
static void ifm_win_details_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_details(win);
}

/*
 *	Show Interface State callback.
 */
static void ifm_win_show_interface_state_toggle_cb(GtkWidget *widget, gpointer data)
{
	const IFMWinDisplay flag = IFM_WIN_DISPLAY_INTERFACE_STATE;
	IFMWinDisplay flags;
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	flags = win->display;
	if(flags & flag)
		flags &= ~flag;
	else
		flags |= flag;
	ifm_win_set_display(
		win,
		flags
	);
}

/*
 *	Show Traffic History callback.
 */
static void ifm_win_show_traffic_history_toggle_cb(GtkWidget *widget, gpointer data)
{
	const IFMWinDisplay flag = IFM_WIN_DISPLAY_TRAFFIC_HISTORY;
	IFMWinDisplay flags;
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	flags = win->display;
	if(flags & flag)
		flags &= ~flag;
	else
		flags |= flag;
	ifm_win_set_display(
		win,
		flags
	);
}

/*
 *	Show Interface State callback.
 */
static void ifm_win_show_traffic_gauges_toggle_cb(GtkWidget *widget, gpointer data)
{
	const IFMWinDisplay flag = IFM_WIN_DISPLAY_TRAFFIC_GAUGES;
	IFMWinDisplay flags;
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	flags = win->display;
	if(flags & flag)
		flags &= ~flag;
	else
		flags |= flag;
	ifm_win_set_display(
		win,
		flags
	);
}

/*
 *      Gauge Style Continuous callback.
 */
static void ifm_win_gauge_style_continuous_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_set_gauge_style(win, IFM_WIN_GAUGE_STYLE_CONTINUOUS);
}

/*
 *      Gauge Style Discrete callback.
 */
static void ifm_win_gauge_style_leds_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_set_gauge_style(win, IFM_WIN_GAUGE_STYLE_LEDS);
}

/*
 *      Gauge Style Led callback.
 */
static void ifm_win_gauge_style_gradient_leds_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_set_gauge_style(win, IFM_WIN_GAUGE_STYLE_GRADIENT_LEDS);
}

/*
 *	Close callback.
 */
static void ifm_win_close_cb(GtkWidget *widget, gpointer data)
{
	IFMWin *win = IFM_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	ifm_win_unmap(win);
}


/*
 *	Connect.
 */
static void ifm_win_connect(IFMWin *win)
{
	gint pid;
	const gchar *cmd;

	win->freeze_count++;

	/* Get the connect command */
	cmd = win->connect_cmd;
	if(STRISEMPTY(cmd))
	{
		win->freeze_count--;
		return;
	}

	/* Already connected? */
	if(win->if_state == IFM_WIN_INTERFACE_STATE_UP)
	{
		win->freeze_count--;
		return;
	}

	/* Execute the connect command */
	pid = process_execute(cmd);
	if(pid > -1)
	{
		/* Mark that we initiated a connect */
		win->flags |= IFM_WIN_CONNECT_PENDING;
	}
	else
	{
		/* Execute failed */
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to execute:\n\
\n\
    %s\n\
\n\
%s.\n",
			cmd,
			g_strerror(error_code)
		);
		CDialogSetTransientFor(win->toplevel);
		(void)CDialogGetResponse(
			"Connect Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
	}

	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Disconnect.
 */
static void ifm_win_disconnect(IFMWin *win)
{
	gint pid;
	const gchar *cmd;

	win->freeze_count++;

	/* Get the disconnect command */
	cmd = win->disconnect_cmd;
	if(STRISEMPTY(cmd))
	{
		win->freeze_count--;
		return;
	}

	/* Already disconnected? */
	if(win->if_state == IFM_WIN_INTERFACE_STATE_DOWN)
	{
		win->freeze_count--;
		return;
	}

	/* Execute the disconnect command */
	pid = process_execute(cmd);
	if(pid > -1)
	{
		/* Remove any connect pending flags */
		win->flags &= ~IFM_WIN_CONNECT_PENDING;
	}
	else
	{
		/* Execute failed */
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to execute:\n\
\n\
    %s\n\
\n\
%s.\n",
			cmd,
			g_strerror(error_code)
		);
		CDialogSetTransientFor(win->toplevel);
		(void)CDialogGetResponse(
			"Disconnect Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
	}

	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Reconnect.
 */
static void ifm_win_reconnect(IFMWin *win)
{
	gint pid;
	const gchar *cmd;

	win->freeze_count++;

	/* Get the reconnect command */
	cmd = win->reconnect_cmd;
	if(STRISEMPTY(cmd))
	{
		win->freeze_count--;
		return;
	}

	/* Execute the reconnect command */
	pid = process_execute(cmd);
	if(pid > -1)
	{
		/* Mark that we initiated a reconnect by setting the
		 * connect pending flag, this flag should not get
		 * removed until the interface state goes to down and
		 * then back to up, meaning that when the interface
		 * state is down this flag will remain set and the
		 * interface state will be displayed as connecting
		 * while it is down
		 */
		win->flags |= IFM_WIN_CONNECT_PENDING;
	}
	else
	{
		/* Execute failed */
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to execute:\n\
\n\
    %s\n\
\n\
%s.\n",
			cmd,
			g_strerror(error_code)
		);
		CDialogSetTransientFor(win->toplevel);
		(void)CDialogGetResponse(
			"Reconnect Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
	}

	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Details.
 *
 *	Creates, sets up, and maps the Details IFMPropDlg.
 */
static void ifm_win_details(IFMWin *win)
{
	IFMPropDlg *d;

	win->freeze_count++;

	/* Get the Details IFMPropDlg, creating it as needed */
	d = win->prop_dlg;
	if(d == NULL)
		win->prop_dlg = d = ifm_prop_dlg_new(win->last_prop_dlg_geometry);
	if(d != NULL)
	{
		GList *route_table;

		/* Set the initial values to the Details IFMPropDlg */
		ifm_prop_dlg_set_interface_name(
			d,
			win->if_name
		);
		ifm_prop_dlg_set_stats(
			d,
			&win->if_stats
		);
		ifm_prop_dlg_set_state(
			d,
			(win->if_state == IFM_WIN_INTERFACE_STATE_UP) ? TRUE : FALSE
		);
		ifm_prop_dlg_set_interface_configuration(
			d,
			&win->if_config
		);
		ifm_prop_dlg_set_con_info(
			d,
			win->chat_report_path,
			&win->con_info,
			win->bandwidth_bytes_per_sec,
			win->user_set_bandwidth
		);
		ifm_prop_dlg_route_table_clear(d);
		route_table = route_table_list();
		if(route_table != NULL)
		{
			GList *glist;
			for(glist = route_table;
			    glist != NULL;
			    glist = g_list_next(glist)
			)
				ifm_prop_dlg_route_table_append(
					d,
					ROUTE_ITEM(glist->data)
				);
			route_table_list_delete(route_table);
		}

		/* Map the Details IFMPropDlg */
		ifm_prop_dlg_map(d);
	}

	win->freeze_count--;
}


/*
 *	Creates a new IFMon Window.
 */
IFMWin *ifm_win_new(
	const gchar *if_name,
	const gulong bandwidth_bytes_per_sec,
	const gboolean user_set_bandwidth,
	const GtkOrientation orientation,
	const gboolean show_border,
	const gboolean show_title,
	const gboolean animate_wm_icon,
	const IFMWinDisplay display,
	const IFMWinGaugeStyle gauge_style,
	const gint gauge_length,
	const gint gauge_nticks,
	const gulong stats_update_int_ms,
	const gchar *chat_report_path
)
{
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*toplevel;
	IFMWin *win = IFM_WIN(g_malloc0(sizeof(IFMWin)));
	if(win == NULL)
		return(NULL);

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_new(): win=%p: if_name=\"%s\"\n", win, if_name);
#endif

	win->accelgrp = accelgrp = gtk_accel_group_new();
	win->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#if 0
	win->busy_count = 0;
	win->freeze_count = 0;
#endif
	win->flags =	(show_border ? IFM_WIN_SHOW_BORDER : 0) |
			(show_title ? IFM_WIN_SHOW_TITLE : 0) |
			(animate_wm_icon ? IFM_WIN_ANIMATE_WM_ICON : 0);
	win->orientation = orientation;
	win->display = display;
	win->gauge_style = gauge_style;
	win->gauge_length = gauge_length;
	win->gauge_nticks = gauge_nticks;
#if 0
	win->last_wm_icon_pixmap = NULL;
	win->last_state_icon_pixmap = NULL;
#endif

	win->if_name = STRDUP(if_name);
#if 0
	(void)if_get_config(
		if_name,
		&win->if_config
	);
#endif
	win->bandwidth_bytes_per_sec = MAX(
		bandwidth_bytes_per_sec,
		IFM_MIN_BANDWIDTH
	);
	win->user_set_bandwidth = user_set_bandwidth;

	win->if_state = IFM_WIN_INTERFACE_STATE_UNKNOWN;
#if 0
	(void)if_get_con_info_from_chat_report(
		chat_report_path,
		if_name,
		&win->con_info
	);
#endif
	win->chat_report_path = STRDUP(chat_report_path);

	win->stats_update_int_ms = MAX(
		stats_update_int_ms,
		IFM_WIN_MIN_STATS_UPDATE_INTERVAL
	);
	/* Initialize our IFStats to the current values so that we get
	 * more accurate calculations in ifm_win_update_stats()
	 */
	(void)if_stat(
		if_name,
		&win->if_stats
	);
	win->stat_timer = g_timer_new();
	g_timer_start(win->stat_timer);		/* Start timer now since
						 * we just got the IFStats */

	win->freeze_count++;

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_object_set_data_full(
		GTK_OBJECT(w), IFM_WIN_KEY,
		win, ifm_win_data_destroy_cb
	);
	gtk_window_set_wmclass(
		GTK_WINDOW(w),
		IFM_WIN_WM_WINDOW_NAME,		/* Window name */
		IFM_WM_CLASS_NAME		/* Class name */
	);
	gtk_widget_set_name(w, IFM_WIN_TOPLEVEL_WIDGET_NAME);
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	if(if_name != NULL)
		gtk_window_set_title(GTK_WINDOW(w), if_name);
	gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_FOCUS_CHANGE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "realize",
		GTK_SIGNAL_FUNC(ifm_win_realize_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "delete_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "focus_in_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "focus_out_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(ifm_win_event_cb), win
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);

	/* Call ifm_win_realize_cb to create the rest of the IFMWin */
	gtk_widget_realize(w);

	win->freeze_count--;

	return(win);
}

/*
 *	Sets the IFMon Window's display flags.
 */
void ifm_win_set_display(
	IFMWin *win,
	const IFMWinDisplay display
)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	if(win->display == display)
		return;

	win->freeze_count++;

	win->display = display;

	/* Interface State */
	w = win->state_toplevel;
	if(w != NULL)
	{
		if(display & IFM_WIN_DISPLAY_INTERFACE_STATE)
			gtk_widget_show(w);
		else
			gtk_widget_hide(w);
	}

	/* Traffic History */
	w = win->traffic_history_toplevel;
	if(w != NULL)
	{
		if(display & IFM_WIN_DISPLAY_TRAFFIC_HISTORY)
			gtk_widget_show(w);
		else
			gtk_widget_hide(w);
	}

	/* Traffic Gauges */
	w = win->traffic_gauges_toplevel;
	if(w != NULL)
	{
		if(display & IFM_WIN_DISPLAY_TRAFFIC_GAUGES)
			gtk_widget_show(w);
		else
			gtk_widget_hide(w);
	}

	/* Update the values displayed on the GtkWidgets */
	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Sets the gauge style.
 */
void ifm_win_set_gauge_style(
	IFMWin *win,
	const IFMWinGaugeStyle gauge_style
)
{
	if(win == NULL)
		return;

	if(win->gauge_style == gauge_style)
		return;

	win->freeze_count++;

	win->gauge_style = gauge_style;

	/* Update the styles of the gauges */
	ifm_win_gauge_set_gauge_style(
		win->tx_traffic_gauge,
		gauge_style
	);
	ifm_win_gauge_set_gauge_style(
		win->rx_traffic_gauge,
		gauge_style
	);

	/* Update the values displayed on the GtkWidgets */
	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Sets the connect command.
 */
void ifm_win_set_connect_command(
	IFMWin *win,
	const gchar *cmd
)
{
	if(win == NULL)
		return;

	if(win->connect_cmd == cmd)
		return;

	win->freeze_count++;

	g_free(win->connect_cmd);
	win->connect_cmd = STRDUP(cmd);

	/* Update the values displayed on the GtkWidgets */
	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Sets the disconnect command.
 */
void ifm_win_set_disconnect_command(
	IFMWin *win,
	const gchar *cmd
)
{
	if(win == NULL)
		return;

	if(win->disconnect_cmd == cmd)
		return;

	win->freeze_count++;

	g_free(win->disconnect_cmd);
	win->disconnect_cmd = STRDUP(cmd);

	/* Update the values displayed on the GtkWidgets */
	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Sets the reconnect command.
 */
void ifm_win_set_reconnect_command(
	IFMWin *win,
	const gchar *cmd
)
{
	if(win == NULL)
		return;

	if(win->reconnect_cmd == cmd)
		return;

	win->freeze_count++;

	g_free(win->reconnect_cmd);
	win->reconnect_cmd = STRDUP(cmd);

	/* Update the values displayed on the GtkWidgets */
	ifm_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Gets the current interface statistics and redraw the display as
 *	needed to reflect any changes in values.
 *
 *	The IFMWin's timer will be checked and restarted in order to
 *	determine the time elapsed between calls to this function.
 *
 *	The IFMWin's interface statistics will be updated with the
 *	current values.
 */
void ifm_win_update_stats(IFMWin *win)
{
	gulong		cur_bytes, last_bytes,
			bandwidth_bytes_int;
	gfloat		tx_coeff,
			rx_coeff;
	gdouble elapsed_sec;
	IFStat stat_buf;
	IFMWinStateIconCodes state_icon_code;

	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 3)
g_print("ifm_win_update_stats() %p\n", win);
#endif

	win->freeze_count++;

	/* Get the interface's current statistics */
	(void)if_stat(
		win->if_name,
		&stat_buf
	);

	/* Get the elapsed time and restart the timer immediately after
	 * getting the interface's statistics so we better isolate the
	 * cycle origin and timming calculations relative to this point
	 * here
	 */
	elapsed_sec = g_timer_elapsed(
		win->stat_timer,
		NULL
	);
	g_timer_start(win->stat_timer);

	/* Calculate the bandwidth in bytes based on the time elapsed
	 * from the previous call to if_stat() to now with a minimum
	 * allowance of one byte to pass through so that we can use
	 * this to calculate the congestion of traffic to be displayed
	 * on the traffic IFMWinGauges on this cycle
	 */
	bandwidth_bytes_int = (gulong)(
		win->bandwidth_bytes_per_sec * elapsed_sec
	);
	if(bandwidth_bytes_int == 0l)
		bandwidth_bytes_int = 1l;

	/* Calculate the traffic per this interval as a coefficient */
	cur_bytes = stat_buf.tx_bytes;
	last_bytes = win->if_stats.tx_bytes;
	if(cur_bytes > last_bytes)
		tx_coeff = (gfloat)(cur_bytes - last_bytes) / (gfloat)bandwidth_bytes_int;
	else
		tx_coeff = 0.0f;

	cur_bytes = stat_buf.rx_bytes;
	last_bytes = win->if_stats.rx_bytes;
	if(cur_bytes > last_bytes)
		rx_coeff = (gfloat)(cur_bytes - last_bytes) / (gfloat)bandwidth_bytes_int;
	else
		rx_coeff = 0.0f;

	/* Update the Traffic History IFMWinGraph only if we are
	 * displaying it
	 */
	if(win->display & IFM_WIN_DISPLAY_TRAFFIC_HISTORY)
		ifm_win_graph_record(
			win->traffic_history_graph,
			elapsed_sec,
			MIN(tx_coeff + rx_coeff, 1.0f)
		);

	/* Update the Traffic IFMWinGauges */
	ifm_win_gauge_set_value(
		win->tx_traffic_gauge,
		tx_coeff
	);
	ifm_win_gauge_set_value(
		win->rx_traffic_gauge,
		rx_coeff
	);

	/* Update the Details IFMPropDlg's Traffic History IFMWinGraph */
	ifm_prop_dlg_record_history(
		win->prop_dlg,
		elapsed_sec,
		tx_coeff,
		rx_coeff
	);

	/* Copy the new interface statistics obtained from if_stat()
	 * to the IFMWin's interface statistics buffer which will be
	 * used in subsequent displays and calculations in the next
	 * call to this function
	 */
	(void)memcpy(
		&win->if_stats,
		&stat_buf,
		sizeof(IFStat)
	);

	/* Update the displayed state icon and WM icon to accurately
	 * reflect the current values
	 */
	state_icon_code = ifm_win_get_state_icon_code(win);
	ifm_win_set_state_icons_by_code(
		win,
		state_icon_code
	);

	/* Update the interface state GtkButton tip */
	ifm_win_update_interface_state_tip(win);

	/* Update the traffic history IFMWinGraph tip */
	ifm_win_update_traffic_history_tip(win);

	/* Update the statistics on the Details IFMPropDlg */
	ifm_prop_dlg_set_stats(
		win->prop_dlg,
		&win->if_stats
	);

	win->freeze_count--;
}

/*
 *	Checks for any change in the interface's state and updates
 *	the displayed values on the GtkWidgets.
 */
void ifm_win_update_state(IFMWin *win)
{
	gboolean value_changed;
	IFMWinInterfaceState if_state;

	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 3)
g_print("ifm_win_update_state() %p\n", win);
#endif

	win->freeze_count++;

	value_changed = FALSE;

	/* If the user specified a chat(8) report file path then get
	 * the connection information
	 */
	if(!STRISEMPTY(win->chat_report_path))
	{
		IFConInfo con_info_buf;

		/* Get the connection information from the chat(8)
		 * report file
		 */
		if(if_get_con_info_from_chat_report(
			win->chat_report_path,
			win->if_name,
			&con_info_buf
		))
		{
			const gint error_code = (gint)errno;
			if(error_code == ENOENT)
			{
				/* Report file does not exist or no
				 * connection information found in it
				 */

			}
			else
			{

			}

			/* Check for any change in the connection
			 * values
			 */
			if((win->con_info.speed_bits_per_sec != 0l) ||
			   (win->con_info.start_time_sec != 0l)
			)
			{
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_update_state() %p: Connection information changed.\n", win);
#endif
				value_changed = TRUE;
			}

			/* Clear the connection information to indicate
			 * that it is not available
			 */
			(void)memset(
				&win->con_info,
				0x00,
				sizeof(IFConInfo)
			);
		}
		else
		{
			/* Got the connection information
			 *
			 * Update the bandwidth only if the user did
			 * not explicitly set it
			 */
			if(!win->user_set_bandwidth)
			{
				const gulong last_bandwidth_bytes_per_sec = win->bandwidth_bytes_per_sec;
				if(con_info_buf.speed_bits_per_sec != 0l)
				{
					win->bandwidth_bytes_per_sec = con_info_buf.speed_bits_per_sec / 8l;
					if(win->bandwidth_bytes_per_sec < IFM_MIN_BANDWIDTH)
						win->bandwidth_bytes_per_sec = IFM_MIN_BANDWIDTH;
				}
				if(win->bandwidth_bytes_per_sec != last_bandwidth_bytes_per_sec)
					value_changed = TRUE;
			}

			/* Check for any change in the connection
			 * values
			 */
			if((con_info_buf.speed_bits_per_sec != win->con_info.speed_bits_per_sec) ||
			   (con_info_buf.start_time_sec != win->con_info.start_time_sec)
			)
			{
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_update_state() %p: Connection information changed.\n", win);
#endif
				value_changed = TRUE;
			}

			/* Copy the new connection information over to
			 * the current connection information buffer
			 */
			(void)memcpy(
				&win->con_info,
				&con_info_buf,
				sizeof(IFConInfo)
			);
		}
	}

	/* Get the interface's state */
	if(if_get_state(win->if_name))
		if_state = IFM_WIN_INTERFACE_STATE_UP;
	else
		if_state = IFM_WIN_INTERFACE_STATE_DOWN;

	/* Change in interface state? */
	if(if_state != win->if_state)
	{
		IFConfig if_config;

		/* Set the new interface state, (re)get the interface's
		 * address in case it has changed, and mark that we
		 * need to update the displayed values
		 */
		win->if_state = if_state;
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_update_state() %p: win->if_state changed to %i.\n", win, if_state);
#endif
		(void)if_get_config(
			win->if_name,
			&if_config
		);
		(void)memcpy(
			&win->if_config,
			&if_config,
			sizeof(IFConfig)
		);

		/* Check if we initiated a connect or reconnect */
		if(win->flags & IFM_WIN_CONNECT_PENDING)
		{
			/* Check if the interface state is now up */
			if(win->if_state == IFM_WIN_INTERFACE_STATE_UP)
			{
				/* Mark that we are no longer waiting
				 * for a connect or reconnect
				 */
				win->flags &= ~IFM_WIN_CONNECT_PENDING;
			}
			else
			{
				/* Do not remove the connect pending
				 * flag when the interface state goes
				 * from up to down since we may have
				 * initiated a reconnect and need the
				 * interface state to be displayed as
				 * such
				 */
			}
		}

		value_changed = TRUE;
	}

	/* Do we need to update any displayed values due to any changes? */
	if(value_changed)
	{
		const IFMWinStateIconCodes state_icon_code = ifm_win_get_state_icon_code(win);
		IFMPropDlg *d = win->prop_dlg;

		/* Update the displayed state icon and WM icon to
		 * accurately reflect the current values
		 */
		ifm_win_set_state_icons_by_code(
			win,
			state_icon_code
		);

		/* Update the values displayed on the GtkWidgets */
		ifm_win_update_display(win);

		/* Update the Details IFMPropDlg */
		if(d != NULL)
		{
			GList *route_table;

			ifm_prop_dlg_set_state(
				d,
				(win->if_state == IFM_WIN_INTERFACE_STATE_UP) ? TRUE : FALSE
			);
			ifm_prop_dlg_set_interface_configuration(
				d,
				&win->if_config
			);
			ifm_prop_dlg_set_con_info(
				d,
				win->chat_report_path,
				&win->con_info,
				win->bandwidth_bytes_per_sec,
				win->user_set_bandwidth
			);

			ifm_prop_dlg_route_table_clear(d);
			route_table = route_table_list();
			if(route_table != NULL)
			{
				GList *glist;
				for(glist = route_table;
				    glist != NULL;
				    glist = g_list_next(glist)
				)
					ifm_prop_dlg_route_table_append(
						d,
						ROUTE_ITEM(glist->data)
					);
				route_table_list_delete(route_table);
			}

		}
	}

	win->freeze_count--;
}

/*
 *	Gets the appropriate IFMWinStateIconCodes based on the
 *	current values on the IFMWin.
 */
IFMWinStateIconCodes ifm_win_get_state_icon_code(IFMWin *win)
{
	if(win == NULL)
		return(IFM_WIN_INTERFACE_STATE_DOWN);

/* TODO check if connecting */
/*
	return(IFM_WIN_STATE_ICON_CONNECTING);
 */
	/* Get the appropriate pixmap to display on the WM icon based
	 * on the interface state
	 */
	if(win->if_state == IFM_WIN_INTERFACE_STATE_UP)
	{
		/* The interface is up, determine a more specific
		 * pixmap based on the current values of the traffic
		 * gauges
		 */
		const gfloat	min_traffic_coeff = 0.10f,
				tx_coeff = ifm_win_gauge_get_value(win->tx_traffic_gauge),
				rx_coeff = ifm_win_gauge_get_value(win->rx_traffic_gauge);
		if((tx_coeff < min_traffic_coeff) &&
		   (rx_coeff < min_traffic_coeff)
		)
			return(IFM_WIN_STATE_ICON_UP);
		else if((tx_coeff >= min_traffic_coeff) &&
		        (rx_coeff >= min_traffic_coeff)
		)
			return(IFM_WIN_STATE_ICON_UP_RX_TX);
		else if(rx_coeff >= min_traffic_coeff)
			return(IFM_WIN_STATE_ICON_UP_RX);
		else
			return(IFM_WIN_STATE_ICON_UP_TX);
	}
	else
	{
		if(win->flags & IFM_WIN_CONNECT_PENDING)
			return(IFM_WIN_STATE_ICON_CONNECTING);
		else
			return(IFM_WIN_STATE_ICON_DOWN);
	}

	/* Never reached */
	return(IFM_WIN_STATE_ICON_DOWN);
}

/*
 *	Sets the state icon and WM icon by IFMWinStateIconCodes.
 *
 *	The state_icon_code specifies the IFMWinStateIconCodes that
 *	will be used to look up and get the icon graphics to be set.
 *	If the icon(s) are already set to the specified icon graphics
 *	then nothing will be done for the icon in question.
 */
void ifm_win_set_state_icons_by_code(
	IFMWin *win,
	const IFMWinStateIconCodes state_icon_code
)
{
	IFMWinPixmap *p;

	if((win == NULL) ||
	   (state_icon_code < 0) || (state_icon_code >= IFM_WIN_NSTATE_ICONS)
	)
		return;

#if (IFM_DEBUG_LEVEL >= 3)
g_print("ifm_win_set_wm_icon_by_code() %p\n", win);
#endif

	win->freeze_count++;

	/* State icon */
	p = win->state_icon_pixmaps[state_icon_code];
	if(ifm_win_pixmap_is_loaded(p) && (p != win->last_state_icon_pixmap))
	{
		GtkWidget *w = win->state_icon_pm;
		if(w != NULL)
		{
			gtk_pixmap_set(
				GTK_PIXMAP(w),
				p->pixmap, p->mask
			);
			win->last_state_icon_pixmap = p;
		}
	}

	/* Details IFMProp State icon */
	if(win->prop_dlg != NULL)
	{
		p = win->prop_dlg_state_icon_pixmaps[state_icon_code];
		if(ifm_win_pixmap_is_loaded(p))
			ifm_prop_dlg_set_state_icon(
				win->prop_dlg,
				p
			);
	}

	/* WM icon */
	p = win->wm_icon_pixmaps[state_icon_code];
	if((win->flags & IFM_WIN_ANIMATE_WM_ICON) &&
	   ifm_win_pixmap_is_loaded(p) && (p != win->last_wm_icon_pixmap)
	)
	{
		GtkWidget *w = win->toplevel;
		GdkWindow *window = w->window;
		if(window != NULL)
		{
			gdk_window_set_icon(
				window,
				window,
				p->pixmap, p->mask
			);
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_set_state_icons_by_code(): win=%p: WM icon changed to state_icon_code=%i\n", win, state_icon_code);
#endif
			gdk_flush();
			win->last_wm_icon_pixmap = p;
		}
	}
	if(win->flags & IFM_WIN_ANIMATE_WM_ICON)
		ifm_prop_dlg_set_wm_icon(
			win->prop_dlg,
			p
		);

	win->freeze_count--;
}

/*
 *	Formats a time duration string.
 */
static gchar *ifm_win_format_duration_string(const gulong duration_sec)
{
	const gulong	days = duration_sec / (24l * 60l * 60l),
			hours = (duration_sec / (60l * 60l)) % 24l,
			minutes = (duration_sec / 60l) % 60l,
			seconds = duration_sec % 60l;
	if(days > 0l)
		return(g_strdup_printf(
			"%ld %s %.2ld:%.2ld:%.2ld",
			days,
			(days == 1l) ? "day" : "days",
			hours, minutes, seconds
		));
	else
		return(g_strdup_printf(
			"%.2ld:%.2ld:%.2ld",
			hours, minutes, seconds
		));
}

/*
 *	Formats the tip string.
 */
static gchar *ifm_win_format_tip_string(IFMWin *win)
{
	const gulong cur_sec = (gulong)time(NULL);
	IFStat *stat_buf = &win->if_stats;
	IFConInfo *con_info = &win->con_info;
	const IFMWinInterfaceState if_state = win->if_state;
	INAddress *inet_addr = &win->if_config.inet_addr;
	gchar	*duration_s = (con_info->start_time_sec > 0l) ?
		ifm_win_format_duration_string(
			cur_sec - con_info->start_time_sec
		) : g_strdup(""),
		*inet_addr_s = g_strdup_printf(
			"%d.%d.%d.%d",
			inet_addr->u8[0],
			inet_addr->u8[1],
			inet_addr->u8[2],
			inet_addr->u8[3]
		),
		*speed_s = ((if_state == IFM_WIN_INTERFACE_STATE_UP) &&
		(con_info->speed_bits_per_sec > 0l)) ?
			g_strdup_printf("\nConnection Speed: %.1f kbps", con_info->speed_bits_per_sec / 1000.0f) :
			g_strdup(""),
		*s = g_strdup_printf(
"%s: %s %s\n\
Address: %s%s\n\
TX: %ld bytes   %ld packets\n\
RX: %ld bytes   %ld packets",
		win->if_name,
		(if_state == IFM_WIN_INTERFACE_STATE_UP) ? "UP" : "DOWN",
		duration_s,
		inet_addr_s,
		speed_s,
		stat_buf->tx_bytes,
		stat_buf->tx_packets,
		stat_buf->rx_bytes,
		stat_buf->rx_packets
	);
	g_free(duration_s);
	g_free(inet_addr_s);
	g_free(speed_s);

	return(s);
}

/*
 *	Updates the interface state tip.
 */
static void ifm_win_update_interface_state_tip(IFMWin *win)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	/* Only update if the traffic history's static tip is suppose to
	 * be shown
	 */
	if(!(win->flags & IFM_WIN_SHOW_INTERFACE_STATE_TIP))
		return;

	win->freeze_count++;

	w = win->state_btn;
	if(w != NULL)
	{
		gchar *s = ifm_win_format_tip_string(win);
		StaticTipSet(
			w,
			s,
			STATIC_TIP_ALIGN_WIDGET_CENTER, STATIC_TIP_ALIGN_WIDGET_MAX,
			0, 0
		);
		g_free(s);
	}

	win->freeze_count--;
}

/*
 *	Updates the traffic history tip.
 */
static void ifm_win_update_traffic_history_tip(IFMWin *win)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	/* Only update if the traffic history's static tip is suppose to
	 * be shown
	 */
	if(!(win->flags & IFM_WIN_SHOW_TRAFFIC_HISTORY_TIP))
		return;

	win->freeze_count++;

	w = ifm_win_graph_get_drawing_area(win->traffic_history_graph);
	if(w != NULL)
	{
		gchar *s = ifm_win_format_tip_string(win);
		StaticTipSet(
			w,
			s,
			STATIC_TIP_ALIGN_WIDGET_CENTER, STATIC_TIP_ALIGN_WIDGET_MAX,
			0, 0
		);
		g_free(s);
	}

	win->freeze_count--;
}

/*
 *	Updates the displayed values on the widgets.
 */
void ifm_win_update_display(IFMWin *win)
{
	gboolean b;
	gchar *s;
	IFMWinFlags flags;
	IFMWinDisplay display;
	IFMWinGaugeStyle gauge_style;
	IFMWinInterfaceState if_state;
#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_update_display(): win=%p\n", win);
#endif

	if(win == NULL)
		return;

	win->freeze_count++;

	flags = win->flags;
	display = win->display;
	gauge_style = win->gauge_style;
	if_state = win->if_state;

	/* Update the title */
	if(win->if_state == IFM_WIN_INTERFACE_STATE_UP)
	{
		if(win->bandwidth_bytes_per_sec != 0l)
			s = g_strdup_printf(
				"%s - Up - %.1f kbps",
				win->if_name,
				(gfloat)win->bandwidth_bytes_per_sec * 8.0f / 1000.0f
			);
		else
			s = g_strdup_printf(
				"%s - Up",
				win->if_name
			);
	}
	else
	{
		s = g_strdup_printf(
			"%s - Down",
			win->if_name
		);
	}
	gtk_window_set_title(
		GTK_WINDOW(win->toplevel),
		(s != NULL) ? s : ""
	);
	g_free(s);

#define SET_CHECK_MENU_ITEM(_w_,_b_)		\
{ if((_w_) != NULL) {				\
 if(GTK_CHECK_MENU_ITEM(_w_)->active != (_b_))	\
  gtk_check_menu_item_set_active(		\
   GTK_CHECK_MENU_ITEM(_w_), (_b_)		\
  );						\
} }

	/* Details */
	b = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->state_menu_details_mi, b);
	GTK_WIDGET_SET_SENSITIVE(win->details_mi, b);

	/* Connect */
	b = (!STRISEMPTY(win->connect_cmd) &&
	     (if_state != IFM_WIN_INTERFACE_STATE_UP) &&
	     !(flags & IFM_WIN_CONNECT_PENDING)
	) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->state_menu_connect_mi, b);
	GTK_WIDGET_SET_SENSITIVE(win->connect_mi, b);

	/* Disconnect */
	b = (!STRISEMPTY(win->disconnect_cmd) &&
	     ((if_state == IFM_WIN_INTERFACE_STATE_UP) ||
	      (flags & IFM_WIN_CONNECT_PENDING)
	     )
	) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->state_menu_disconnect_mi, b);
	GTK_WIDGET_SET_SENSITIVE(win->disconnect_mi, b);

	/* Reconnect */
	b = (!STRISEMPTY(win->reconnect_cmd) &&
	     (if_state == IFM_WIN_INTERFACE_STATE_UP)
	) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->state_menu_reconnect_mi, b);
	GTK_WIDGET_SET_SENSITIVE(win->reconnect_mi, b);

	/* Display */
	b = (display & IFM_WIN_DISPLAY_INTERFACE_STATE) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->show_interface_state_micheck, b);
	b = (display & IFM_WIN_DISPLAY_TRAFFIC_HISTORY) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->show_traffic_history_micheck, b);
	b = (display & IFM_WIN_DISPLAY_TRAFFIC_GAUGES) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->show_traffic_gauges_micheck, b);

	/* Gauge Style (take the value from the Outgoing Traffic Gauge) */
	b = (gauge_style == IFM_WIN_GAUGE_STYLE_CONTINUOUS) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->gauge_style_continuous_micheck, b);
	b = (gauge_style == IFM_WIN_GAUGE_STYLE_LEDS) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->gauge_style_leds_micheck, b);
	b = (gauge_style == IFM_WIN_GAUGE_STYLE_GRADIENT_LEDS) ? TRUE : FALSE;
	SET_CHECK_MENU_ITEM(win->gauge_style_gradient_leds_micheck, b);

	/* Close */
	b = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->state_menu_close_mi, b);
	GTK_WIDGET_SET_SENSITIVE(win->close_mi, b);

#undef SET_CHECK_MENU_ITEM

	win->freeze_count--;
}

/*
 *	Checks if the IFMWin is mapped.
 */
gboolean ifm_win_is_mapped(IFMWin *win)
{
	if(win == NULL)
		return(FALSE);

	return((win->flags & IFM_WIN_MAPPED) ? TRUE : FALSE);
}

/*
 *	Maps the IFMWin.
 */
void ifm_win_map(IFMWin *win)
{
	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_map() %p\n", win);
#endif

	gtk_widget_show(win->toplevel);
	win->flags |= IFM_WIN_MAPPED;
}

/*
 *	Unmaps the IFMWin.
 */
void ifm_win_unmap(IFMWin *win)
{
	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_unmap() %p\n", win);
#endif
	gtk_widget_hide(win->toplevel);
	win->flags &= ~IFM_WIN_MAPPED;
}

/*
 *	Deletes the IFMWin.
 */
void ifm_win_delete(IFMWin *win)
{
	if(win == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 1)
g_print("ifm_win_delete() %p\n", win);
#endif

	ifm_win_unmap(win);

	/* Destroy the toplevel GtkWindow and call
	 * ifm_win_data_destroy_cb() to delete the rest of the IFMWin
	 */
	GTK_WIDGET_DESTROY(win->toplevel);
}
