#include <gtk/gtk.h>
#include "guiutils.h"
#include "guirgbimg.h"
#include "ifm_win_graph.h"
#include "config.h"


/*
 *      Flags:
 */
typedef enum {
	IFM_WIN_GRAPH_REALIZED	= (1 << 0)
} IFMWinGraphFlags;


/* Callbacks */
static void ifm_win_graph_data_destroy_cb(gpointer data);
static void ifm_win_graph_realize_cb(GtkWidget *widget, gpointer data);
static gint ifm_win_graph_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

/* IFMWinGraph */
IFMWinGraph *ifm_win_graph_new(
	GtkWidget *parent,
	const GtkDirectionType direction,
	const gchar *widget_name,
	const gchar *label_text,
	const gint width, const gint height,
	const gulong capacity_sec
);
GtkWidget *ifm_win_graph_get_drawing_area(IFMWinGraph *graph);
void ifm_win_graph_delete(IFMWinGraph *graph);

/* Resize */
void ifm_win_graph_resize(
	IFMWinGraph *graph,
	const gint width, const gint height
);

/* Draw */
void ifm_win_graph_draw(IFMWinGraph *graph);
void ifm_win_graph_queue_draw(IFMWinGraph *graph);

/* Record */
void ifm_win_graph_record(
	IFMWinGraph *graph,
	const gdouble elapsed_sec,
	const gfloat v
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define IFM_WIN_GRAPH_KEY		"IFMWinGraph"


/*
 *	IFMWinGraph:
 */
struct _IFMWinGraph {
	GtkWidget	*toplevel;		/* GtkFrame */
	gint		freeze_count;
	IFMWinGraphFlags	flags;
	GtkDirectionType	direction;

	gchar		*label_text;
	GtkWidget	*drawing_area;		/* GtkDrawingArea */

	GdkPixmap	*pixmap;		/* Back buffer for drawing_area */

	/* Buffer allocated to the width of drawing_area, each unit
	 * represents a coefficient value from 0.0 to 1.0 mapped to
	 * the full range of the buffer's type of guint8 which is
	 * from 0 to 255 (respectively) */
	guint8		*buf;
	gint		buf_len;

	gulong		capacity_sec;		/* Time capacity of buf */
	gdouble		recorded_capacity_sec;	/* Actual amount of recording */
};


/*
 *	Toplevel GtkObject data "destroy" signal callback.
 */
static void ifm_win_graph_data_destroy_cb(gpointer data)
{
	IFMWinGraph *graph = IFM_WIN_GRAPH(data);
	if(graph == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_graph_data_destroy_cb(): graph=%p\n", graph);
#endif

	/* Handle this signal even if already frozen */

	graph->freeze_count++;

/*	graph->toplevel = NULL; */
	g_free(graph->buf);
	g_free(graph->label_text);
	graph->pixmap = GDK_PIXMAP_UNREF(graph->pixmap);

	graph->freeze_count--;

	g_free(graph);
}

/*
 *	GtkDrawingArea "realize" signal callback.
 */
static void ifm_win_graph_realize_cb(GtkWidget *widget, gpointer data)
{
	GdkWindow *window;
	GtkStyle *style;
	IFMWinGraph *graph = IFM_WIN_GRAPH(data);
	if((widget == NULL) || (graph == NULL))
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_graph_realize_cb(): graph=%p\n", graph);
#endif

	window = widget->window;
	if((window == NULL) || (graph->flags & IFM_WIN_GRAPH_REALIZED))
		return;

	/* Handle this signal even if already frozen */

	graph->freeze_count++;

	/* Set the style's base color as the GdkWindow's background
	 * color
	 */
	style = gtk_widget_get_style(widget);
	if(style != NULL)
		gdk_window_set_background(
			window,
			&style->base[GTK_STATE_NORMAL]
		);

	/* Mark the IFMWinGauge as realized */
	graph->flags |= IFM_WIN_GRAPH_REALIZED;

	graph->freeze_count--;
}

/*
 *	GtkDrawingArea any event signal callback.
 */
static gint ifm_win_graph_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventButton *button;
	IFMWinGraph *graph = IFM_WIN_GRAPH(data);
	if((widget == NULL) || (event == NULL) || (graph == NULL))
		return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
		configure = (GdkEventConfigure *)event;
		ifm_win_graph_resize(
			graph,
			configure->width,
			configure->height
		);
		status = TRUE;
		break;

	  case GDK_EXPOSE:
		expose = (GdkEventExpose *)event;
		ifm_win_graph_draw(graph);
		status = TRUE;
		break;

	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		if(button->button == GDK_BUTTON3)
		{

		}
		break;
	}
	return(status);
}


/*
 *	Creates a new IFMWinGraph.
 */
IFMWinGraph *ifm_win_graph_new(
	GtkWidget *parent,
	const GtkDirectionType direction,
	const gchar *widget_name,
	const gchar *label_text,
	const gint width, const gint height,
	const gulong capacity_sec
)
{
	GtkWidget	*w,
			*parent2;
	IFMWinGraph *graph = IFM_WIN_GRAPH(g_malloc0(sizeof(IFMWinGraph)));
	if(graph == NULL)
		return(NULL);

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_graph_new(): graph=%p: label_text=\"%s\"\n", graph, label_text);
#endif

	graph->direction = direction;
	graph->label_text = STRDUP(label_text);
	graph->capacity_sec = MAX(capacity_sec, 1l);

	graph->freeze_count++;

	/* Toplevel GtkFrame */
	graph->toplevel = w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_object_set_data_full(
		GTK_OBJECT(w), IFM_WIN_GRAPH_KEY,
		graph, ifm_win_graph_data_destroy_cb
	);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkDrawingArea */
	graph->drawing_area = w = gtk_drawing_area_new();
	if(widget_name != NULL)
		gtk_widget_set_name(w, widget_name);
	gtk_widget_set_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "realize",
		GTK_SIGNAL_FUNC(ifm_win_graph_realize_cb), graph
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(ifm_win_graph_event_cb), graph
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(ifm_win_graph_event_cb), graph
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(ifm_win_graph_event_cb), graph
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(ifm_win_graph_event_cb), graph
	);
	gtk_widget_set_usize(
		w,
		width, height
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);

	graph->freeze_count--;

	return(graph);
}

/*
 *	Gets the IFMWinGraph's GtkDrawingArea.
 */
GtkWidget *ifm_win_graph_get_drawing_area(IFMWinGraph *graph)
{
	if(graph == NULL)
		return(NULL);

	return(graph->drawing_area);
}

/*
 *	Deletes the IFMWinGraph.
 */
void ifm_win_graph_delete(IFMWinGraph *graph)
{
	if(graph == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_graph_delete(): graph=%p\n", graph);
#endif
	/* Call ifm_win_graph_data_destroy_cb() to delete the rest of the
	 * IFMWinGraph
	 */
	GTK_WIDGET_DESTROY(graph->toplevel);
}


/*
 *	Resizes the IFMWinGraph.
 */
void ifm_win_graph_resize(
	IFMWinGraph *graph,
	const gint width, const gint height
)
{
	gint new_buf_len;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_graph_resize(): graph=%p: Size %ix%i.\n", graph, width, height);
#endif

	graph->pixmap = GDK_PIXMAP_UNREF(graph->pixmap);
	if((width > 0) && (height > 0) && (graph->drawing_area != NULL))
	{
		GtkWidget *w = graph->drawing_area;
		GdkWindow *window = w->window;
		if(window != NULL)
			graph->pixmap = gdk_pixmap_new(
				window,
				width, height,
				-1
			);
	}

	/* Reallocate the IFMWinGraph buffer to match the width of the
	 * new size
	 */
	new_buf_len = width;
	if(new_buf_len > 0)
	{
		/* Need to increase the buffer allocation? */
		if(new_buf_len > graph->buf_len)
		{
			/* Allocate a new buffer of increased size and
			 * then copy the smaller old buffer segment to
			 * it and zero the rest of the new buffer
			 */
			guint8 *new_buf = (guint8 *)g_malloc(new_buf_len * sizeof(guint8));
			if(new_buf != NULL)
			{
				if((graph->buf != NULL) && (graph->buf_len > 0))
					(void)memcpy(
						new_buf,
						graph->buf,
						graph->buf_len * sizeof(guint8)
					);
				(void)memset(
					new_buf + graph->buf_len,
					0x00,
					(new_buf_len - graph->buf_len) * sizeof(guint8)
				);

				/* Delete the old buffer and transfer
				 * the new buffer to it
				 */
				g_free(graph->buf);
				graph->buf = new_buf;
				graph->buf_len = new_buf_len;
			}
			else
			{
				g_free(graph->buf);
				graph->buf = NULL;
				graph->buf_len = 0;
			}
		}
		/* Need to decrease the buffer allocation? */
		else if(new_buf_len < graph->buf_len)
		{
			graph->buf = (guint8 *)g_realloc(
				graph->buf,
				new_buf_len * sizeof(guint8)
			);
			if(graph->buf != NULL)
				graph->buf_len = new_buf_len;
			else
				graph->buf_len = 0;
		}
	}
	else
	{
		g_free(graph->buf);
		graph->buf = NULL;
		graph->buf_len = 0;
	}
}

/*
 *	Draws the IFMWinGraph.
 */
void ifm_win_graph_draw(IFMWinGraph *graph)
{
	const gint border_minor = 2;
	gint width, height;
	GdkWindow *window;
	GdkDrawable *drawable;
	GtkStateType state;
	GtkWidget *w;
	GtkStyle *style;
	GtkDirectionType direction;

	if(graph == NULL)
		return;

	w = graph->drawing_area;
	if(!GTK_WIDGET_VISIBLE(w))
		return;

	window = w->window;
	if(window == NULL)
		return;

	drawable = (GdkDrawable *)graph->pixmap;
	if(drawable == NULL)
		drawable = (GdkDrawable *)window;

	gdk_window_get_size((GdkWindow *)drawable, &width, &height);
	if((width <= 0) || (height <= 0))
		return;

	direction = graph->direction;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);

	/* Draw the background */
	gdk_draw_rectangle(
		drawable,
		style->base_gc[state],
		TRUE,
		0, 0,
		width, height
	);

	/* Draw the graph */
	if((graph->buf != NULL) && (graph->buf_len > 0))
	{
		gint x;
		const gint buf_len = graph->buf_len;
		const guint8	*buf = graph->buf,
				*buf_ptr, *buf_end;
		gfloat coeff;
		GdkGC *gc = style->bg_gc[GTK_STATE_SELECTED];

		if(graph->direction == GTK_DIR_LEFT)
		{
			/* Draw each unit as 1:1 for each pixel along
			 * the GdkDrawable's X axis
			 */
			for(x = width - 1,
			    buf_ptr = buf,
			    buf_end = buf + buf_len;
			    buf_ptr < buf_end;
			    x--,
			    buf_ptr++
			)
			{
				coeff = (gfloat)*buf_ptr / (gfloat)0xFF;
				gdk_draw_line(
					drawable,
					gc,
					x,
					height - 1,
					x,
					height - ((height - 1) * coeff) - 1
				);
			}
		}
		else if(graph->direction == GTK_DIR_RIGHT)
		{
			/* Draw each unit as 1:1 for each pixel along
			 * the GdkDrawable's X axis
			 */
			for(x = 0,
			    buf_ptr = buf,
			    buf_end = buf + buf_len;
			    buf_ptr < buf_end;
			    x++,
			    buf_ptr++
			)
			{
				coeff = (gfloat)*buf_ptr / (gfloat)0xFF;
				gdk_draw_line(
					drawable,
					gc,
					x,
					height - 1,
					x,
					height - ((height - 1) * coeff) - 1
				);
			}
		}
	}

	/* Draw the grids */
	if(width >= 50)
	{
		gint x, y;
		GdkGC *gc = style->fg_gc[state];

		/* Horizontal Grids */
		y = (gint)(height * 0.25f);
		gdk_draw_line(
			drawable,
			gc,
			0, y,
			width, y
		);
		y = (gint)(height * 0.5f);
		gdk_draw_line(
			drawable,
			gc,
			0, y,
			width, y
		);
		y = (gint)(height * 0.75f);
		gdk_draw_line(
			drawable,
			gc,
			0, y,
			width, y
		);

		/* Vertical Grids */
		x = (gint)(width * 0.25f);
		gdk_draw_line(
			drawable,
			gc,
			x, 0,
			x, height
		);
		x = (gint)(width * 0.5f);
		gdk_draw_line(
			drawable,
			gc,
			x, 0,
			x, height
		);
		x = (gint)(width * 0.75f);
		gdk_draw_line(
			drawable,
			gc,
			x, 0,
			x, height
		);

		/* Draw text */
		if(style->font != NULL)
		{
			const gdouble recorded_capacity_sec = graph->recorded_capacity_sec;
			GdkFont *font = style->font;
			const gint font_height = font->ascent + font->descent;

			/* Main Label Text */
			if(graph->label_text != NULL)
			{
				const gchar *s = graph->label_text;
		                GdkTextBounds b;
		                gdk_string_bounds(font, s, &b);
				if(direction == GTK_DIR_LEFT)
					x = 0 + border_minor;
				else
					x = width - b.width - border_minor;
				y = border_minor;
	        	        gdk_draw_string(
		                        drawable,
	        	                font,
	                	        style->text_gc[GTK_STATE_SELECTED],
		                        x - b.lbearing,
					y + font->ascent,
					s
				);
			}

			/* Time labels for vertical grid lines */
			if((recorded_capacity_sec > 0.0) && (width >= 100))
			{
				gchar *s = g_strdup_printf(
					"%.1fs",
					(gfloat)recorded_capacity_sec
				);
				if(s != NULL)
				{
		        	        GdkTextBounds b;
			                gdk_string_bounds(font, s, &b);
					if(direction == GTK_DIR_LEFT)
						x = 0 + border_minor;
					else
						x = width - b.width - border_minor;
					y = height - border_minor - font_height;
	        		        gdk_draw_string(
		        	                drawable,
	        	        	        font,
		                	        style->text_gc[state],
			                        x - b.lbearing,
						y + font->ascent,
						s
					);
					g_free(s);
				}
				s = g_strdup_printf(
					"%.1fs",
					(gfloat)recorded_capacity_sec * 0.5f
				);
				if(s != NULL)
				{
		        	        GdkTextBounds b;
			                gdk_string_bounds(font, s, &b);
					x = (width * 0.5f) + border_minor;
					y = height - border_minor - font_height;
	        		        gdk_draw_string(
		        	                drawable,
	        	        	        font,
		                	        style->text_gc[state],
			                        x - b.lbearing,
						y + font->ascent,
						s
					);
					g_free(s);
				}
			}
		}
	}

	/* Send the drawable to the window if it is not the window */
	if(drawable != (GdkDrawable *)window)
		gdk_draw_pixmap(
			(GdkDrawable *)window,
			style->white_gc,
			drawable,
			0, 0,
			0, 0,
			width, height
		);
}

/*
 *	Queues the IFMWinGraph to draw.
 */
void ifm_win_graph_queue_draw(IFMWinGraph *graph)
{
	if(graph == NULL)
		return;

	gtk_widget_queue_draw(graph->drawing_area);
}


/*
 *	Records the value as a coefficient.
 */
void ifm_win_graph_record(
	IFMWinGraph *graph,
	const gdouble elapsed_sec,
	const gfloat v
)
{
	const guint8 v8 = (guint8)(0xFF * CLIP(v, 0.0f, 1.0f));

	if(graph == NULL)
		return;

	/* Buffer not allocated? */
	if((graph->buf == NULL) || (graph->buf_len <= 0))
		return;

	/* Record the value relative to a specified time capacity? */
	if(graph->capacity_sec > 0l)
	{
		const gint buf_len = graph->buf_len;
		guint8	*buf = graph->buf,
			*buf_src_ptr,
			*buf_tar_ptr;
		const gint last_v = (gint)buf[0];
		gint	cur_v,
			step_v;
		const gdouble capacity_sec = (gdouble)graph->capacity_sec;

		/* Calculate the number of buffer units we need to fill
		 * for the elapsed time
		 */
		gint dbuf_len_int = (gint)(
			(((gdouble)graph->buf_len / capacity_sec) *
			elapsed_sec) +
			0.5f			/* For better rounding-off */
		);
		if(dbuf_len_int < 1)
			dbuf_len_int = 1;

		/* Shift the buffer to make room for the new units */
		for(buf_tar_ptr = buf + buf_len - 1,
		    buf_src_ptr = buf_tar_ptr - dbuf_len_int;
		    buf_src_ptr >= buf;
		    buf_tar_ptr--,
		    buf_src_ptr--
		)
			*buf_tar_ptr = *buf_src_ptr;

		/* Set the new units */
		step_v = ((gint)v8 - last_v) / dbuf_len_int;
		cur_v = last_v + step_v;
		for(buf_tar_ptr = buf + MIN(dbuf_len_int, buf_len) - 1;
		    buf_tar_ptr >= buf;
		    buf_tar_ptr--
		)
		{
			*buf_tar_ptr = (guint8)cur_v;
			cur_v += step_v;
		}

		/* Update the actual amount of time recorded */
		graph->recorded_capacity_sec = ((gdouble)buf_len /
			(gdouble)dbuf_len_int) * elapsed_sec;
	}
	else
	{
		/* Record a fixed one unit per call and ignore the
		 * elapsed time
		 */
		const gint buf_len = graph->buf_len;
		guint8	*buf = graph->buf,
			*buf_src_ptr,
			*buf_tar_ptr;

		/* Shift the buffer to make room for the new units */
		for(buf_tar_ptr = buf + buf_len - 1,
		    buf_src_ptr = buf_tar_ptr - 1;
		    buf_src_ptr >= buf;
		    buf_tar_ptr--,
		    buf_src_ptr--
		)
			*buf_tar_ptr = *buf_src_ptr;

		/* Set the new value */
		buf[0] = v8;

		/* Update the actual amount of time recorded */
		graph->recorded_capacity_sec = elapsed_sec * (gdouble)buf_len;
	}


	/* Check if we have recorded */


	ifm_win_graph_queue_draw(graph);
}
