#include <stdlib.h>
#include <signal.h>
#include <gtk/gtk.h>
#include "guiutils.h"
#include "cdialog.h"
#include "if_list.h"
#include "if_config.h"
#include "ifm_cb.h"
#include "ifm_win.h"
#include "ifm_core.h"
#include "config.h"


static gint ifm(const gint argc, const gchar **argv);

static gint ifm_list(const gint argc, const gchar **argv);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	IFMon.
 */
static gint ifm(const gint argc, const gchar **argv)
{
	gboolean	initialized_gtk = FALSE,
			user_set_bandwidth = FALSE,
			show_border = TRUE,
			show_title = TRUE,
			animate_wm_icon = TRUE;
	gint		i,
			gauge_length = IFM_WIN_DEF_GAUGE_LENGTH,
			gauge_nticks = IFM_WIN_DEF_GAUGE_TICKS;
	const gchar	*arg,
			*interface_name = IFM_DEF_INTERFACE_NAME,
			*chat_report_path = NULL,
			*connect_command = NULL,
			*disconnect_command = NULL,
			*reconnect_command = NULL;
	gulong		bandwidth_bytes_per_sec = IFM_DEF_BANDWIDTH,
			stats_update_interval_ms = IFM_WIN_DEF_STATS_UPDATE_INTERVAL;
	GtkOrientation orientation = GTK_ORIENTATION_HORIZONTAL;
	IFMWinDisplay display =	IFM_WIN_DISPLAY_TRAFFIC_GAUGES |
				IFM_WIN_DISPLAY_INTERFACE_STATE;
	IFMWinGaugeStyle gauge_style = IFM_WIN_DEF_GAUGE_STYLE;
	IFMWin *win = NULL;
	IFMCore *core = NULL;

#define CLEANUP_RETURN(_v_)	{		\
						\
 if(core != NULL) {				\
  if(core->ifm_wins_list != NULL) {		\
   g_list_foreach(				\
    core->ifm_wins_list,			\
    (GFunc)ifm_win_delete,			\
    NULL					\
   );						\
   g_list_free(core->ifm_wins_list);		\
  }						\
  g_free(core);					\
 }						\
						\
 if(initialized_gtk) {				\
  /* Shut down all dialogs */			\
  CDialogShutdown();				\
						\
  initialized_gtk = FALSE;			\
 }						\
						\
 return(_v_);					\
}

	/* Set up the UNIX signal callbacks */
#ifdef SIGINT
	signal(SIGINT, ifm_signal_cb);
#endif
#ifdef SIGTERM
	signal(SIGTERM, ifm_signal_cb);
#endif

	/* Allocate the core */
	core = IFM_CORE(g_malloc0(sizeof(IFMCore)));
	if(core == NULL)
	{
		CLEANUP_RETURN(-3);
	}

	/* Handle the command line arguments */
	for(i = 1; i < argc; i++)
	{
		arg = argv[i];
		if(arg == NULL)
			continue;

		/* Bandwidth */
		if(!g_strcasecmp(arg, "--bandwidth") ||
		   !g_strcasecmp(arg, "-bandwidth") ||
		   !g_strcasecmp(arg, "--speed") ||
		   !g_strcasecmp(arg, "-speed") ||
		   !strcmp(arg, "--b") ||
		   !strcmp(arg, "-b")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				user_set_bandwidth = TRUE;
				bandwidth_bytes_per_sec = (gulong)atol((const char *)arg);
				if(bandwidth_bytes_per_sec < IFM_MIN_BANDWIDTH)
				{
					g_printerr(
"%s: Specified bandwidth %ld bytes per second too small, using minimum bandwidth %i bytes per second.\n",
						argv[i - 1],
						bandwidth_bytes_per_sec,
						IFM_MIN_BANDWIDTH
					);
					bandwidth_bytes_per_sec = IFM_MIN_BANDWIDTH;
				}
			}
		}
		/* Bandwidth in bits per second */
		else if(!strcmp(arg, "--B") ||
			!strcmp(arg, "-B")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				user_set_bandwidth = TRUE;
				bandwidth_bytes_per_sec = (gulong)atol((const char *)arg) / 8l;
				if(bandwidth_bytes_per_sec < IFM_MIN_BANDWIDTH)
				{
					g_printerr(
"%s: Specified bandwidth %ld bytes per second too small, using minimum bandwidth %i bytes per second.\n",
						argv[i - 1],
						bandwidth_bytes_per_sec,
						IFM_MIN_BANDWIDTH
					);
					bandwidth_bytes_per_sec = IFM_MIN_BANDWIDTH;
				}
			}
		}
		/* No Border */
		else if(!g_strcasecmp(arg, "--no-border") ||
		        !g_strcasecmp(arg, "-no-border") ||
		        !g_strcasecmp(arg, "--no-borders") ||
		        !g_strcasecmp(arg, "-no-borders") ||
		        !strcmp(arg, "--w") ||
		        !strcmp(arg, "-w")
		)
		{
			show_border = FALSE;
		}
		/* No Title Bar */
		else if(!g_strcasecmp(arg, "--no-title-bar") ||
		        !g_strcasecmp(arg, "--no-title") ||
		        !g_strcasecmp(arg, "-no-title-bar") ||
		        !g_strcasecmp(arg, "-no-title") ||
		        !strcmp(arg, "--t") ||
		        !strcmp(arg, "-t")
		)
		{
			show_title = FALSE;
		}
		/* Static WM Icon */
		else if(!g_strcasecmp(arg, "--static-wm-icon") ||
		        !g_strcasecmp(arg, "-static-wm-icon") ||
		        !strcmp(arg, "--y") ||
		        !strcmp(arg, "-y")
		)
		{
			animate_wm_icon = FALSE;
		}
		/* Display */
		else if(!strcmp(arg, "--d") ||
		        !strcmp(arg, "-d")
		)
		{
			display = 0;		/* Reset the display */
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				gchar **strv = g_strsplit(
					arg,
					" ",
					-1
				);
				if(strv != NULL)
				{
					gint j;
					const gchar *s;
					for(j = 0; strv[j] != NULL; j++)
					{
						s = strv[j];
						if(!g_strcasecmp(s, "state"))
							display |= IFM_WIN_DISPLAY_INTERFACE_STATE;
						else if(!g_strcasecmp(s, "traffic_history_graph") ||
							!g_strcasecmp(s, "traffic-history-graph") ||
							!g_strcasecmp(s, "traffic_history") ||
							!g_strcasecmp(s, "traffic-history")
						)
							display |= IFM_WIN_DISPLAY_TRAFFIC_HISTORY;
						else if(!g_strcasecmp(s, "traffic_gauges") ||
                		                        !g_strcasecmp(s, "traffic-gauges") ||
                		                        !g_strcasecmp(s, "trafficgauges") ||
                		                        !g_strcasecmp(s, "gauges")
						)
							display |= IFM_WIN_DISPLAY_TRAFFIC_GAUGES;
						else
							g_printerr(
"%s: Warning: Unsupported display value \"%s\".\n",
								argv[j - 1],
								s
							);
					}
					g_strfreev(strv);
				}
			}
		}
		/* Gauge Style */
		else if(!g_strcasecmp(arg, "--gauge-style") ||
			!g_strcasecmp(arg, "-gauge-style") ||
			!strcmp(arg, "-s")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				if(!g_strcasecmp(arg, "continuous"))
					gauge_style = IFM_WIN_GAUGE_STYLE_CONTINUOUS;
				else if(!g_strcasecmp(arg, "leds"))
					gauge_style = IFM_WIN_GAUGE_STYLE_LEDS;
				else if(!g_strcasecmp(arg, "gradient_leds"))
					gauge_style = IFM_WIN_GAUGE_STYLE_GRADIENT_LEDS;
				else
					g_printerr(
"%s: Warning: Unsupported gauge style \"%s\".\n",
						argv[i - 1],
						arg
					);
			}
		}
		/* Orientation */
		else if(!g_strcasecmp(arg, "--orientation") ||
			!g_strcasecmp(arg, "-orientation") ||
			!strcmp(arg, "-o")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				if(*arg == 'h')
					orientation = GTK_ORIENTATION_HORIZONTAL;
				else if(*arg == 'v')
					orientation = GTK_ORIENTATION_VERTICAL;
				else
					g_printerr(
"%s: Warning: Unsupported argument \"%s\".\n",
						argv[i - 1], arg
					);
			}
		}
		/* Bar Length */
		else if(!g_strcasecmp(arg, "--length") ||
			!g_strcasecmp(arg, "-length") ||
			!strcmp(arg, "-l")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				gauge_length = (gint)atoi((const char *)arg);
				if(gauge_length < 1)
					gauge_length = 1;
			}
		}
		/* Bar Ticks */
		else if(!g_strcasecmp(arg, "--ticks") ||
			!g_strcasecmp(arg, "-ticks") ||
			!strcmp(arg, "--n") ||
			!strcmp(arg, "-n")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				gauge_nticks = (gint)atoi((const char *)arg);
				if(gauge_nticks < 1)
					gauge_nticks = 1;
			}
		}
		/* Stats Update Interval */
		else if(!g_strcasecmp(arg, "--interval") ||
			!g_strcasecmp(arg, "-interval") ||
			!strcmp(arg, "-i")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				stats_update_interval_ms = (gulong)atol((const char *)arg);
				if(stats_update_interval_ms < IFM_WIN_MIN_STATS_UPDATE_INTERVAL)
				{
					g_printerr(
"%s: Specified statistics update interval %ld milliseconds too small,\
 using minimum statistics update interval %i milliseconds.\n",
						argv[i - 1],
						stats_update_interval_ms,
						IFM_WIN_MIN_STATS_UPDATE_INTERVAL
					);
					stats_update_interval_ms = IFM_WIN_MIN_STATS_UPDATE_INTERVAL;
				}
			}
		}
		/* Report Path */
		else if(!g_strcasecmp(arg, "--report-path") ||
			!g_strcasecmp(arg, "-report-path") ||
			!g_strcasecmp(arg, "--report") ||
			!g_strcasecmp(arg, "-report") ||
			!strcmp(arg, "--r") ||
			!strcmp(arg, "-r")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				chat_report_path = arg;
			}
		}
		/* Connect Command */
		else if(!g_strcasecmp(arg, "--connect-command") ||
			!g_strcasecmp(arg, "-connect-command") ||
			!g_strcasecmp(arg, "--connect") ||
			!g_strcasecmp(arg, "-connect")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				connect_command = arg;
			}
		}
		/* Disconnect Command */
		else if(!g_strcasecmp(arg, "--disconnect-command") ||
			!g_strcasecmp(arg, "-disconnect-command") ||
			!g_strcasecmp(arg, "--disconnect") ||
			!g_strcasecmp(arg, "-disconnect")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				disconnect_command = arg;
			}
		}
		/* Reconnect Command */
		else if(!g_strcasecmp(arg, "--reconnect-command") ||
			!g_strcasecmp(arg, "-reconnect-command") ||
			!g_strcasecmp(arg, "--reconnect") ||
			!g_strcasecmp(arg, "-reconnect")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(STRISEMPTY(arg))
			{
				g_printerr(
"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
			else
			{
				reconnect_command = arg;
			}
		}
		/* Skip these arguments so that gtk_window_apply_args()
		 * handles them
		 */
		else if(gtk_is_window_arg(arg))
		{
			i++;
		}
		/* Single character argument? */
		else if((*arg == '-') ? (arg[1] != '-') : FALSE)
		{
			const gchar *v = arg + 1;

			while(*v != '\0')
			{
#if 0
				switch(*v)
				{
				    case 's':
					auto_start = TRUE;
					break;
				    default:
					g_printerr(
"-%c: Unsupported argument.\n",
						c
					);
					CLEANUP_RETURN(-2);
					break;
				}
#endif
				v++;
			}
		}
		/* Unsupported argument */
		else if(*arg == '-')
		{
			g_printerr(
"%s: Unsupported argument.\n",
				arg
			);
			CLEANUP_RETURN(-2);
		}
		/* Interface Name */
		else
		{
			interface_name = arg;
		}
	}


	/* Initialize GTK as needed */
	if(!initialized_gtk)
	{
		if(!gtk_init_check((gint *)&argc, (gchar ***)&argv))
		{
			g_printerr("Unable to initialize GTK.\n");
			CLEANUP_RETURN(-1);
		}
		initialized_gtk = TRUE;

		/* Initialize GDK RGB buffers system */
		gdk_rgb_init();

		CDialogInit();
	}

	/* Create the IFMWin */
	win = ifm_win_new(
		interface_name,
		bandwidth_bytes_per_sec,
		user_set_bandwidth,
		orientation,
		show_border,
		show_title,
		animate_wm_icon,
		display,
		gauge_style,
		gauge_length,
		gauge_nticks,
		stats_update_interval_ms,
		chat_report_path
	);
	if(win == NULL)
	{
		CLEANUP_RETURN(-1);
	}
	core->ifm_wins_list = g_list_append(
		core->ifm_wins_list,
		win
	);
	gtk_window_apply_args(
		GTK_WINDOW(win->toplevel),
		(gint)argc, (gchar **)argv
	);
	if(!STRISEMPTY(connect_command))
		ifm_win_set_connect_command(
			win,
			connect_command
		);
	if(!STRISEMPTY(disconnect_command))
		ifm_win_set_disconnect_command(
			win,
			disconnect_command
		);
	if(!STRISEMPTY(reconnect_command))
		ifm_win_set_reconnect_command(
			win,
			reconnect_command
		);
	ifm_win_map(win);


	/* Set the main timeout callback to check when all windows
	 * have been closed and to break out of the main GTK main
	 * loop
	 */
	core->main_timeout_id = gtk_timeout_add(
		1000l,
		ifm_timeout_cb, core
	);

	/* Push the first and main GTK main loop */
	gtk_main();

	/* Delete all the windows, the core, shut down GTK, and
	 * return success
	 */
	CLEANUP_RETURN(0);
#undef CLEANUP_RETURN
}


/*
 *	Print a list of interfaces that are up.
 */
static gint ifm_list(const gint argc, const gchar **argv)
{
	GList *interfaces_list = if_list();
	if(interfaces_list != NULL)
	{
		const gchar *if_name;
		GList *glist;
		for(glist = interfaces_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
			if_name = (const gchar *)glist->data;
			if(if_get_state(if_name))
				g_print(
					"%s\n",
					if_name
				);
			g_free(glist->data);
		}
		g_list_free(interfaces_list);
	}
	return(0);
}


int main(int argc, char *argv[])
{
	gint		i,
			status;
	const gchar *arg;
	gint (*ifm_main_ptr)(const gint, const gchar **) = ifm;

	/* Handle the basic command line arguments */
	for(i = 1; i < argc; i++)
	{
		arg = (const gchar *)argv[i];
		if(arg == NULL)
			continue;

		/* Help */
		if(!g_strcasecmp(arg, "--help") ||
		   !g_strcasecmp(arg, "-help") ||
		   !g_strcasecmp(arg, "--h") ||
		   !g_strcasecmp(arg, "-h") ||
		   !g_strcasecmp(arg, "-?")
		)
		{
			g_print("%s", PROG_HELP_MESG);
			return(0);
		}
		/* Version */
		else if(!g_strcasecmp(arg, "--version") ||
			!g_strcasecmp(arg, "-version")
		)
		{
			g_print(
				"%s",
				PROG_NAME " Version " PROG_VERSION "\n"
				PROG_COPYRIGHT
			);
			return(0);
		}
		/* List */
		else if(!g_strcasecmp(arg, "--list") ||
			!g_strcasecmp(arg, "-list")
		)
		{
			ifm_main_ptr = ifm_list;
		}
	}

	/* Run the selected main function */
	status = ifm_main_ptr((gint)argc, (const gchar **)argv);

	/* Return value must be positive */
	if(status < 0)
		status = -status;

	return(status);
}
